#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Flags all files in a processing pass and its descendants + flags the RAW files
Parameters:
   <Processing pass> : processing pass to start from (can be /RealData)
   <run> : run number
   <flag> : flag to set
"""
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script


def getProcessingPasses(bkDict, headPass):
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    passes = {}
    res = BookkeepingClient().getProcessingPass(bkDict, headPass)
    if not res["OK"]:
        gLogger.error(
            "Cannot load the processing passes for head % in Version %s Data taking condition %s"
            % (headPass, bkDict["ConfigVersion"], bkDict["ConditionDescription"])
        )
        gLogger.error(res["Message"])
        DIRAC.exit(2)

    for recordList in res["Value"]:
        if recordList["TotalRecords"] == 0:
            continue
        parNames = recordList["ParameterNames"]

        found = False
        for thisId in range(len(parNames)):
            parName = parNames[thisId]
            if parName == "Name":
                found = True
                break
        if found:
            for reco in recordList["Records"]:
                recoName = headPass + "/" + reco[0]
                passes[recoName] = True
                passes.update(getProcessingPasses(bkDict, recoName))

    return passes


@Script()
def main():
    Script.setUsageMessage(f"Usage: {Script.scriptName} <Processing Pass> <run> <status> <flag>")
    Script.parseCommandLine()
    args = Script.getPositionalArgs()
    if len(args) < 3:
        Script.showHelp(exitCode=2)

    realData = "/Real Data"
    processing = args[0].replace("/RealData", realData)
    if processing == "":
        processing = realData
    run = int(args[1])
    flag = args[2]

    # gLogger.error('Please wait for the OK from Stefan before actually flagging')
    # gLogger.error('Do ELOG the flag in the DQ ELOG fr future rerefence.')
    # DIRAC.exit(2)

    #
    # Processing pass needs to start as "/Real Data" for FULL stream flagging
    #

    if realData not in processing:
        print("You forgot /Real Data in the processing pass:  ", processing)
        DIRAC.exit(2)
    #
    # Make sure it is a known processing pass
    #

    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient

    bkClient = BookkeepingClient()
    res = bkClient.getRunInformations(run)
    if not res["OK"]:
        gLogger.error("Cannot load the information for run %d" % run)
        gLogger.error(res["Message"])
        DIRAC.exit(2)

    dtd = res["Value"]["DataTakingDescription"]
    configName = res["Value"]["Configuration Name"]
    configVersion = res["Value"]["Configuration Version"]

    bkDict = {"ConfigName": configName, "ConfigVersion": configVersion, "ConditionDescription": dtd}

    if processing != realData:
        knownPasses = getProcessingPasses(bkDict, "")
        if processing not in knownPasses:
            gLogger.error(f"{processing} is not a valid processing pass.")
            DIRAC.exit(2)

    recoPasses = getProcessingPasses(bkDict, processing)
    if realData in recoPasses:
        recoPasses.pop(realData)

    # Flag the run realData first

    res = bkClient.setRunAndProcessingPassDataQuality(run, realData, flag)

    if not res["OK"]:
        print(res["Message"])
        DIRAC.exit(2)
    else:
        print("Run %d RAW files flagged %s" % (run, flag))

    # Now the reconstruction and stripping processing passes
    for thisPass in recoPasses:
        res = bkClient.setRunAndProcessingPassDataQuality(run, thisPass, flag)

        if not res["OK"]:
            print(res["Message"])
            DIRAC.exit(2)
        else:
            print("Run %d Processing Pass %s flagged %s" % (run, thisPass, flag))

    DIRAC.exit()


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_Resources_Catalog_RAWIntegrityClient."""
from unittest import mock
import unittest

from six.moves import reload_module
from DIRAC import gLogger
import LHCbDIRAC.Resources.Catalog.RAWIntegrityClient as moduleTested

################################################################################


class RAWIntegrityClient_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""

        gLogger.setLevel("DEBUG")

        # Mock external libraries / modules not interesting for the unit test
        mock_pathFinder = mock.MagicMock()
        mock_pathFinder.getServiceURL.return_value = "cookiesURL"
        self.mock_pathFinder = mock_pathFinder

        mock_RPC = mock.MagicMock()
        mock_RPC.addFile.return_value = {"OK": True}
        #    mock_RPC.addMigratingReplicas.return_value    = { 'OK' : True }
        #    mock_RPC.removeMigratingFiles.return_value    = { 'OK' : True }
        #    mock_RPC.removeMigratingReplicas.return_value = { 'OK' : True }
        #
        mock_RPCClient = mock.MagicMock()
        mock_RPCClient.return_value = mock_RPC
        self.mock_RPCClient = mock_RPCClient

        # Add mocks to moduleTested
        moduleTested.PathFinder = self.mock_pathFinder
        moduleTested.RPCClient = self.mock_RPCClient

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.RAWIntegrityClient

    def tearDown(self):
        """TearDown."""
        del self.testClass
        del self.moduleTested
        del self.mock_pathFinder
        del self.mock_RPCClient


################################################################################


class RAWIntegrityClient_Success(RAWIntegrityClient_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""

        catalog = self.testClass()
        self.assertEqual("RAWIntegrityClient", catalog.__class__.__name__)

    def test_exists(self):
        """tests the output of exists."""

        catalog = self.testClass()

        res = catalog.exists("1")
        self.assertTrue(res["OK"])

        res = catalog.exists({})
        self.assertFalse(res["OK"])

        res = catalog.exists(["path1"])
        self.assertTrue(res["OK"])

        res = catalog.exists({"A": 1, "B": 2})
        self.assertTrue(res["OK"])
        self.assertEqual({"Failed": {}, "Successful": {"A": False, "B": False}}, res["Value"])

    def test_addFile(self):
        """tests the output of addFile."""

        catalog = self.testClass()
        catalog.rawIntegritySrv = mock.MagicMock()

        res = catalog.addFile({"1": {"PFN": "pfn", "Size": 123, "SE": "aSe", "GUID": "aGuid", "Checksum": "aCksm"}})
        self.assertTrue(res["OK"])

        fileDict = {"PFN": "pfn", "Size": "10", "SE": "se", "GUID": "guid", "Checksum": "checksum"}

        fileDict["Size"] = "10"

        #    res = catalog.addFile( { 'lfn1' : fileDict } )
        #    self.assertTrue(res['OK'])
        #    self.assertEqual( { 'Successful' : { 'lfn1' : True }, 'Failed' : {} }, res['Value'] )
        #
        #    res = catalog.addFile( { 'lfn1' : fileDict, 'lfn2' : fileDict } )
        #    self.assertTrue(res['OK'])
        #    self.assertEqual( { 'Successful' : { 'lfn1' : True, 'lfn2' : True }, 'Failed' : {} }, res['Value'] )
        #
        #    mock_RPC = mock.Mock()
        #    mock_RPC.addFile.return_value = { 'OK' : False, 'Message' : 'Bo!' }
        #
        #    self.moduleTested.RPCClient.return_value = mock_RPC
        #    catalog = self.testClass()
        #
        #    res = catalog.addFile( { 'lfn1' : fileDict } )
        #    self.assertTrue(res['OK'])
        #    self.assertEqual( { 'Successful' : {}, 'Failed' : {'lfn1' : 'Bo!' } }, res['Value'] )
        #
        #    res = catalog.addFile( { 'lfn1' : fileDict, 'lfn2' : fileDict } )
        #    self.assertTrue(res['OK'])
        #    self.assertEqual( { 'Successful' : {}, 'Failed' : {  'lfn1' : 'Bo!', 'lfn2' : 'Bo!' } }, res['Value'] )

        # Restore the module
        self.moduleTested.RPCClient.return_value = self.mock_RPCClient
        reload_module(self.moduleTested)


################################################################################



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""This LHCbDIRAC agent takes BkQueries from the TranformationDB, and issue a
query to the BKK, for populating a table in the Transformation DB, with all the
files in input to a transformation.

A pickle file is used as a cache.
"""
import os
import time
import datetime
import pickle
import queue as Queue

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.ThreadPool import ThreadPool
from DIRAC.Core.Utilities.ThreadSafe import Synchronizer
from DIRAC.Core.Utilities.List import breakListIntoChunks
from DIRAC.Core.Utilities.ReturnValues import returnValueOrRaise, SErrorException
from DIRAC.TransformationSystem.Agent.TransformationAgentsUtilities import TransformationAgentsUtilities
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = "Transformation/BookkeepingWatchAgent"
gSynchro = Synchronizer()


class BookkeepingWatchAgent(AgentModule, TransformationAgentsUtilities):
    """LHCbDIRAC only agent.

    A threaded agent.
    """

    def __init__(self, *args, **kwargs):
        """c'tor."""
        AgentModule.__init__(self, *args, **kwargs)
        TransformationAgentsUtilities.__init__(self)

        self.bkQueriesToBeChecked = Queue.Queue()
        self.bkQueriesInCheck = []

        self.fullUpdatePeriod = 86400
        self.bkUpdateLatency = 7200
        self.debug = False

        self.transInThread = {}

        self.pickleFile = "BookkeepingWatchAgent.pkl"
        self.chunkSize = 1000

        # No need to give full list as it is in the CS anyway
        self.pluginsWithRunInfo = []

        self.timeLog = {}
        self.fullTimeLog = {}
        self.bkQueries = {}
        self.redo = set()

        self.transClient = None
        self.bkClient = None

    def initialize(self):
        """Make the necessary initializations.

        The ThreadPool is created here, the _execute() method is what each
        thread will execute.
        """

        self.fullUpdatePeriod = self.am_getOption("FullUpdatePeriod", self.fullUpdatePeriod)
        self.bkUpdateLatency = self.am_getOption("BKUpdateLatency", self.bkUpdateLatency)
        self.debug = self.am_getOption("verbose", self.debug)

        self.pickleFile = os.path.join(self.am_getWorkDirectory(), self.pickleFile)
        self.chunkSize = self.am_getOption("maxFilesPerChunk", self.chunkSize)

        self.pluginsWithRunInfo = Operations().getValue(
            "TransformationPlugins/PluginsWithRunInfo", self.pluginsWithRunInfo
        )

        self._logInfo(f"Full Update Period: {self.fullUpdatePeriod} seconds")
        self._logInfo(f"BK update latency : {self.bkUpdateLatency} seconds")
        self._logInfo(f"Plugins with run info: {', '.join(self.pluginsWithRunInfo)}")

        self.transClient = TransformationClient()
        self.bkClient = BookkeepingClient()

        try:
            with open(self.pickleFile, "rb") as pf:
                self.timeLog = pickle.load(pf)
                self.fullTimeLog = pickle.load(pf)
                self.bkQueries = pickle.load(pf)
                self.redo = pickle.load(pf)
            self._logInfo("successfully loaded Log from", self.pickleFile, "initialize")
        except (EOFError, OSError):
            self._logInfo("failed loading Log from", self.pickleFile, "initialize")
            self.timeLog = {}
            self.fullTimeLog = {}
            self.bkQueries = {}
            self.redo = set()

        maxNumberOfThreads = self.am_getOption("maxThreadsInPool", 1)
        threadPool = ThreadPool(maxNumberOfThreads, maxNumberOfThreads)

        for i in range(maxNumberOfThreads):
            threadPool.generateJobAndQueueIt(self._execute, [i])

        return S_OK()

    @gSynchro
    def __dumpLog(self):
        """dump the log in the pickle file."""
        if self.pickleFile:
            try:
                with open(self.pickleFile, "wb") as pf:
                    pickle.dump(self.timeLog, pf)
                    pickle.dump(self.fullTimeLog, pf)
                    pickle.dump(self.bkQueries, pf)
                    pickle.dump(self.redo, pf)
                self._logVerbose(f"successfully dumped Log into {self.pickleFile}")
            except OSError as e:
                self._logError(f"fail to open {self.pickleFile}: {e}")
            except pickle.PickleError as e:
                self._logError(f"fail to dump {self.pickleFile}: {e}")
            except ValueError as e:
                self._logError(f"fail to close {self.pickleFile}: {e}")

    ################################################################################

    def execute(self):
        """Main execution method.

        Just fills a list, and a queue, with BKKQueries ID.
        """

        # Get all the transformations
        result = self.transClient.getTransformations(condDict={"Status": ["Active", "Idle"]})
        if not result["OK"]:
            self._logError("Failed to get transformations.", result["Message"])
            return S_OK()
        transIDsList = [int(transDict["TransformationID"]) for transDict in result["Value"]]
        res = self.transClient.getTransformationsWithBkQueries(transIDsList)
        if not res["OK"]:
            self._logError("Failed to get transformations with Bk Queries.", res["Message"])
            return S_OK()
        transIDsWithBkQueriesList = res["Value"]

        _count = 0
        # Process each transformation
        for transID in transIDsWithBkQueriesList:
            if transID in self.bkQueriesInCheck:
                continue
            self.bkQueriesInCheck.append(transID)
            self.bkQueriesToBeChecked.put(transID)
            _count += 1

        self._logInfo("Out of %d transformations, %d put in thread queue" % (len(result["Value"]), _count))

        self.__dumpLog()
        return S_OK()

    def _execute(self, threadID):
        """Real executor.

        This is what is executed by the single threads - so do not return here! Just continue
        """

        while True:  # not self.bkQueriesToBeChecked.empty():
            transID = None
            startTime = None
            try:
                transID = self.bkQueriesToBeChecked.get()
                self.transInThread[transID] = " [Thread%d] [%s] " % (threadID, str(transID))

                startTime = time.time()
                self._logInfo(f"Processing transformation {transID}.", transID=transID)

                res = self.transClient.getTransformation(transID, extraParams=False)
                if not res["OK"]:
                    self._logError("Failed to get transformation", res["Message"], transID=transID)
                    continue
                transPlugin = res["Value"]["Plugin"]

                res = self.transClient.getBookkeepingQuery(transID)
                if not res["OK"]:
                    self._logError("Failed to get BkQuery", res["Message"], transID=transID)
                    continue
                bkQuery = res["Value"]

                # Determine the correct time stamp to use for this transformation
                now = datetime.datetime.utcnow()
                self.__timeStampForTransformation(transID, bkQuery, now)

                try:
                    filesMetadata = self.__getFiles(transID, bkQuery, now)
                except (RuntimeError, SErrorException) as e:
                    # In case we failed a full query, we should retry full query until successful
                    if "StartDate" not in bkQuery:
                        self.bkQueries.pop(transID, None)
                    self._logError(f"Failed to get response from the Bookkeeping: {e}", "", "__getFiles", transID)
                    continue

                runDict = {}

                # There is no need to add the run information for a transformation that doesn't need it
                if transPlugin in self.pluginsWithRunInfo:
                    for lfn, metadata in filesMetadata.items():
                        runID = metadata.get("RunNumber", None)
                        if isinstance(runID, ((str,), (int,))):
                            runDict.setdefault(int(runID), []).append(lfn)
                    try:
                        self.__addRunsMetadata(transID, list(runDict))
                    except RuntimeError as e:
                        self._logException(
                            "Failure adding runs metadata", method="__addRunsMetadata", lException=e, transID=transID
                        )
                else:
                    runDict[None] = list(filesMetadata)

                # Add all new files to the transformation
                for runID in sorted(runDict):
                    lfnList = runDict[runID]
                    # We enter all files of a run at once, otherwise do it by chunks
                    lfnChunks = [lfnList] if runID else breakListIntoChunks(lfnList, self.chunkSize)
                    for lfnChunk in lfnChunks:
                        # Add the files to the transformation
                        self._logVerbose(f"Adding {len(lfnChunk)} lfns for transformation", transID=transID)
                        result = self.transClient.addFilesToTransformation(transID, lfnChunk)
                        if not result["OK"]:
                            self._logError(
                                f"Failed to add {len(lfnChunk)} lfns to transformation",
                                result["Message"],
                                transID=transID,
                            )
                            self.redo.add(transID)
                            continue
                        else:
                            # Handle errors
                            errors = {}
                            for lfn, error in result["Value"]["Failed"].items():
                                errors.setdefault(error, []).append(lfn)
                            for error, lfns in errors.items():
                                self._logWarn("Failed to add files to transformation", error, transID=transID)
                                self._logVerbose("\n\t".join([""] + lfns))
                            # Add the metadata and RunNumber to the newly inserted files
                            addedLfns = [
                                lfn for (lfn, status) in result["Value"]["Successful"].items() if status == "Added"
                            ]
                            if addedLfns:
                                # Add files metadata: size and file type
                                lfnDict = {
                                    lfn: {
                                        "Size": filesMetadata[lfn]["FileSize"],
                                        "FileType": filesMetadata[lfn]["FileType"],
                                    }
                                    for lfn in addedLfns
                                }
                                res = self.transClient.setParameterToTransformationFiles(transID, lfnDict)
                                if not res["OK"]:
                                    self._logError(
                                        "Failed to set transformation files metadata", res["Message"], transID=transID
                                    )
                                    continue
                                # Add run information if it exists
                                if runID:
                                    self._logInfo(
                                        "Added %d files to transformation for run %d, now including run information"
                                        % (len(addedLfns), runID),
                                        transID=transID,
                                    )
                                    self._logVerbose(
                                        "Associating %d files to run %d" % (len(addedLfns), runID), transID=transID
                                    )
                                    res = self.transClient.addTransformationRunFiles(transID, runID, addedLfns)
                                    if not res["OK"]:
                                        self._logError(
                                            "Failed to associate %d files to run %d" % (len(addedLfns), runID),
                                            res["Message"],
                                            transID=transID,
                                        )
                                        continue
                                else:
                                    self._logInfo(f"Added {len(addedLfns)} files to transformation", transID=transID)

            except Exception as x:  # pylint: disable=broad-except
                self._logException(
                    "Exception while adding files to transformation", lException=x, method="_execute", transID=transID
                )
            finally:
                self._logInfo("Processed transformation", transID=transID, reftime=startTime)
                if transID in self.bkQueriesInCheck:
                    self.bkQueriesInCheck.remove(transID)
                self.transInThread.pop(transID, None)

        return S_OK()

    @gSynchro
    def __timeStampForTransformation(self, transID, bkQuery, now):
        """Determine the correct time stamp to use for this transformation."""

        fullTimeLog = self.fullTimeLog.setdefault(transID, now)
        bkQueryLog = self.bkQueries.setdefault(transID, {})

        original_start_date = bkQueryLog.pop("StartDate", None)
        self.bkQueries[transID] = bkQuery.copy()
        if (
            transID in self.timeLog
            and bkQueryLog == bkQuery
            and (now - fullTimeLog) < datetime.timedelta(seconds=self.fullUpdatePeriod)
        ):
            # If it is more than a day since the last reduced query, make a full query just in case
            timeStamp = self.timeLog[transID]
            delta = datetime.timedelta(seconds=self.bkUpdateLatency)
            if transID in self.redo:
                if original_start_date:
                    bkQuery["StartDate"] = original_start_date
                self.redo.remove(transID)
            else:
                bkQuery["StartDate"] = (timeStamp - delta).strftime("%Y-%m-%d %H:%M:%S")
        if "StartDate" not in bkQuery:
            self.fullTimeLog[transID] = now

    def __getFiles(self, transID, bkQuery, now):
        """Perform the query to the Bookkeeping."""
        self._logInfo(f"Using BK query for transformation: {str(bkQuery)}", transID=transID)

        # Make interface compatible between getFiles and getFilesWithMetadata
        if "DataQualityFlag" in bkQuery:
            bkQuery["DataQuality"] = bkQuery["DataQualityFlag"]
        if "ProductionID" in bkQuery:
            bkQuery["Production"] = bkQuery["ProductionID"]
        bkQuery = {"Visible": "ALL"} | bkQuery

        start = time.time()
        bkQuery["OnlyParameters"] = ["FileName", "RunNumber", "FileSize", "FileType"]
        res = returnValueOrRaise(self.bkClient.getFilesWithMetadata(bkQuery))
        records = res["Records"]
        self._logVerbose(f"BK query time: {time.time() - start:.2f} seconds.", transID=transID)

        indexes = res["ParameterNames"]
        fileName_idx = indexes.index("FileName")
        runNumber_idx = indexes.index("RunNumber")
        fileSize_idx = indexes.index("FileSize")
        fileType_idx = indexes.index("FileType")

        lfn_run = {
            rec[fileName_idx]: {
                "RunNumber": rec[runNumber_idx],
                "FileSize": rec[fileSize_idx],
                "FileType": rec[fileType_idx],
            }
            for rec in records
        }

        self.__updateTimeStamp(transID, now)

        self._logInfo(f"Obtained {len(lfn_run)} files from BK", transID=transID)
        return lfn_run

    @gSynchro
    def __updateTimeStamp(self, transID, now):
        """Update time stamp for current transformation to now."""
        self.timeLog[transID] = now

    def __addRunsMetadata(self, transID, runsList):
        """Add the run metadata."""
        runsInCache = self.transClient.getRunsInCache({"Name": ["TCK", "CondDb", "DDDB"]})
        if not runsInCache["OK"]:
            raise RuntimeError(runsInCache["Message"])
        newRuns = list(set(runsList) - set(runsInCache["Value"]))
        if newRuns:
            self._logVerbose(f"Associating run metadata to {len(newRuns)} runs", transID=transID)
            res = self.bkClient.getRunInformation({"RunNumber": newRuns, "Fields": ["TCK", "CondDb", "DDDB"]})
            if not res["OK"]:
                raise RuntimeError(res["Message"])
            else:
                for run, runMeta in res["Value"].items():
                    res = self.transClient.addRunsMetadata(run, runMeta)
                    if not res["OK"]:
                        raise RuntimeError(res["Message"])
        # Add run duration to the metadata
        runsInCache = self.transClient.getRunsInCache({"Name": ["Duration"]})
        if not runsInCache["OK"]:
            raise RuntimeError(runsInCache["Message"])
        newRuns = list(set(runsList) - set(runsInCache["Value"]))
        if newRuns:
            self._logVerbose(f"Associating run duration to {len(newRuns)} runs", transID=transID)
            res = self.bkClient.getRunInformation({"RunNumber": newRuns, "Fields": ["JobStart", "JobEnd"]})
            if not res["OK"]:
                raise RuntimeError(res["Message"])
            else:
                for run, runMeta in res["Value"].items():
                    duration = (runMeta["JobEnd"] - runMeta["JobStart"]).seconds
                    res = self.transClient.addRunsMetadata(run, {"Duration": duration})
                    if not res["OK"]:
                        raise RuntimeError(res["Message"])

    def finalize(self):
        """Gracious finalization."""
        if self.bkQueriesInCheck:
            self._logInfo(f"Wait for queue to get empty before terminating the agent ({len(self.transInThread)} tasks)")
            self.bkQueriesInCheck = []
            while self.transInThread:
                time.sleep(2)
            self.log.info("Threads are empty, terminating the agent...")
        return S_OK()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Remove the given file or a list of files from the File Catalog and from the
storage."""

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()
    dmScript.registerFileSwitches()

    Script.registerSwitch(
        "", "IncludeProcessedFiles", "  Forced to set Removed the files in status Processed (default:not reset)"
    )
    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile] [<LFN>] [<LFN>...]",
            ]
        )
    )

    Script.parseCommandLine()

    from LHCbDIRAC.DataManagementSystem.Client.ScriptExecutors import executeRemoveFiles
    from DIRAC import exit

    exit(executeRemoveFiles(dmScript))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Report info on event types for MC"""
from collections import defaultdict
import DIRAC
from DIRAC.Core.Base.Script import Script
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.BookkeepingSystem.Client.BKQuery import BKQuery


@Script()
def main():
    Script.registerSwitch("", "FileType=", "FileType to search [default:ALLSTREAMS.DST]")

    Script.setUsageMessage(__doc__ + "\n".join(["Usage:", f"  {Script.scriptName} [option] eventType (mandatory)"]))
    fileType = "ALLSTREAMS.DST"
    Script.parseCommandLine(ignoreErrors=True)
    for switch in Script.getUnprocessedSwitches():
        if switch[0] == "FileType":
            fileType = str(switch[1])

    args = Script.getPositionalArgs()
    if len(args) < 1:
        Script.showHelp(exitCode=1)

    eventTypes = args[0]
    bkQuery = BKQuery({"EventType": eventTypes, "ConfigName": "MC"}, fileTypes=fileType, visible=True)
    print("BK query:", bkQuery)
    prods = bkQuery.getBKProductions()

    pathInfo = defaultdict(lambda: defaultdict(int))
    for prod in prods:
        res = BookkeepingClient().getProductionInformation(prod)
        if not res["OK"]:
            print(res["Message"])
            DIRAC.exit(1)
        value = res["Value"]
        path = value["Path"].split("\n")[1]
        nfiles = 0
        for nf in value["Number of files"]:
            if nf[1] == fileType:
                nfiles += nf[0]
        nevents = 0
        for ne in value["Number of events"]:
            if ne[0] == fileType:
                nevents += ne[1]
        pathInfo[path]["Files"] += nfiles
        pathInfo[path]["Events"] += nevents

    for path in sorted(pathInfo):
        print(f"{path}: {pathInfo[path]['Files']} files, {pathInfo[path]['Events']} events")


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""In general for data processing productions we need to completely abandon the
'by hand' reschedule operation such that accidental reschedulings don't result
in data being processed twice.

For all above cases the following procedure should be used to achieve 100%:

- Starting from the data in the Production DB for each transformation
  look for files in the following status:

    - Assigned
    - MaxReset

  some of these will correspond to the final WMS status 'Failed'.

For files in MaxReset and Assigned:

- Discover corresponding job WMS ID
- Check that there are no outstanding requests for the job
  o wait until all are treated before proceeding
- Check that none of the job input data has BK descendants for the current production
  o if the data has a replica flag it means all was uploaded successfully - should be investigated by hand
  o if there is no replica flag can proceed with file removal from FC / storage (can be disabled by flag)
- Mark the recovered input file status as 'Unused' in the ProductionDB if they were not in MaxReset
"""
import datetime

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient

from LHCbDIRAC.DataManagementSystem.Client.ConsistencyChecks import ConsistencyChecks
from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

AGENT_NAME = "Transformation/DataRecoveryAgent"


class DataRecoveryAgent(AgentModule):
    """Standard DIRAC agent class."""

    def __init__(self, *args, **kwargs):
        """c'tor."""
        AgentModule.__init__(self, *args, **kwargs)

        self.transClient = None
        self.reqClient = None
        self.consChecks = None

        self.enableFlag = True
        self.transformationTypes = []
        self.transLogger = self.log

    #############################################################################

    def initialize(self):
        """Sets defaults."""
        self.am_setOption("shifterProxy", "ProductionManager")

        self.transClient = TransformationClient()
        self.reqClient = ReqClient()
        self.consChecks = ConsistencyChecks(interactive=False, transClient=self.transClient)

        transformationTypes = Operations().getValue("Transformations/DataProcessing", [])
        extendableTTypes = Operations().getValue("Transformations/ExtendableTransfTypes", ["MCSimulation"])
        self.transformationTypes = list(set(transformationTypes) - set(extendableTTypes))

        return S_OK()

    #############################################################################
    def execute(self):
        """The main execution method."""
        # Configuration settings
        self.enableFlag = self.am_getOption("EnableFlag", True)
        self.log.verbose(f"Enable flag is {self.enableFlag}")
        if not self.transformationTypes:
            self.log.warn("No transformation types to look for... aborting")
            return S_OK()

        transformationStatus = self.am_getOption("TransformationStatus", ["Active", "Completing"])
        fileSelectionStatus = self.am_getOption("FileSelectionStatus", ["Assigned", "MaxReset"])
        unrecoverableStatus = self.am_getOption("UnrecoverableStatus", ["MaxReset"])
        updateStatus = self.am_getOption("FileUpdateStatus", "Unused")
        wmsStatusList = self.am_getOption("WMSStatus", ["Failed"])

        # only worry about files > 12hrs since last update
        selectDelay = self.am_getOption("SelectionDelay", 1)  # hours

        transformationDict = {}
        for transStatus in transformationStatus:
            result = self.__getEligibleTransformations(transStatus, self.transformationTypes)
            if not result["OK"]:
                self.log.error(
                    "Could not obtain eligible transformations",
                    f"Status '{transStatus}': {result['Message']}",
                )
                return result

            if not result["Value"]:
                self.log.info(
                    'No "%s" transformations of types %s to process.'
                    % (transStatus, ", ".join(self.transformationTypes))
                )
                continue

            transformationDict.update(result["Value"])

        self.log.info(
            f"Selected {len(transformationDict)} transformations of types {', '.join(self.transformationTypes)}"
        )
        self.log.verbose(f"Transformations selected:\n{', '.join(transformationDict)}")

        for transformation, typeName in transformationDict.items():
            self.transLogger = self.log.getSubLogger(f"Trans-{transformation}")
            result = self.__selectTransformationFiles(transformation, fileSelectionStatus)
            if not result["OK"]:
                self.transLogger.error(
                    "Could not select files for transformation", f"{transformation}: {result['Message']}"
                )
                continue
            fileDict = result["Value"]
            if not fileDict:
                self.transLogger.verbose(
                    "No files in status %s selected for transformation %s"
                    % (", ".join(fileSelectionStatus), transformation)
                )
                continue

            title = f"Looking at transformation {transformation}, type {typeName} "
            self.transLogger.info("=" * len(title))
            self.transLogger.info(title)

            self.transLogger.info(f"Selected {len(fileDict)} files with status {','.join(fileSelectionStatus)}")
            result = self.__obtainWMSJobIDs(transformation, fileDict, selectDelay, wmsStatusList)
            if not result["OK"]:
                self.transLogger.error("Could not obtain jobs for files of transformation", result["Message"])
                continue
            jobFileDict = result["Value"]
            if not jobFileDict:
                self.transLogger.info(f"No {' or '.join(wmsStatusList)} jobs found for selected files")
                continue

            self.transLogger.verbose(f"Looking at WMS jobs {','.join(str(jobID) for jobID in jobFileDict)}")

            fileCount = sum(len(lfnList) for lfnList in jobFileDict.values())
            self.transLogger.verbose(
                f"{str(fileCount) if fileCount else 'No'} files are selected after examining WMS jobs"
            )
            if not fileCount:
                continue

            result = self.__removePendingRequestsJobs(jobFileDict)
            if not result["OK"]:
                self.transLogger.error("Error while removing jobs with pending requests", result["Message"])
                continue
            # This method modifies the input dictionary
            if not jobFileDict:
                self.transLogger.info("No WMS jobs without pending requests to process.")
                continue

            fileCount = sum(len(lfnList) for lfnList in jobFileDict.values())
            self.transLogger.info(
                "%s files are selected in %d jobs after removing any job with pending requests"
                % (str(fileCount) if fileCount else "No", len(jobFileDict))
            )
            if not fileCount:
                continue

            jobsThatDidntProduceOutputs, jobsThatProducedOutputs = self.__checkdescendants(transformation, jobFileDict)
            title = f"======== Transformation {transformation}: results ========"
            self.transLogger.info(title)
            self.transLogger.info(f"\tTotal jobs that can be updated now: {len(jobsThatDidntProduceOutputs)}")
            if jobsThatProducedOutputs:
                self.transLogger.info(f"\t{len(jobsThatProducedOutputs)} jobs have descendants")
            else:
                self.transLogger.info("\tNo jobs have descendants")

            filesToUpdate = []
            filesMaxReset = []
            filesWithDescendants = []
            for job, fileList in jobFileDict.items():
                if job in jobsThatDidntProduceOutputs:
                    recoverableFiles = {lfn for lfn in fileList if fileDict[lfn][1] not in unrecoverableStatus}
                    filesToUpdate += list(recoverableFiles)
                    filesMaxReset += list(set(fileList) - recoverableFiles)
                elif job in jobsThatProducedOutputs:
                    filesWithDescendants += fileList

            if filesToUpdate:
                self.transLogger.info(f"\tUpdating {len(filesToUpdate)} files to '{updateStatus}'")
                result = self.__updateFileStatus(transformation, filesToUpdate, updateStatus)
                if not result["OK"]:
                    self.transLogger.error("\tRecoverable files were not updated", result["Message"])

            if filesMaxReset:
                self.transLogger.info(
                    "\t%d files are in %s status and have no descendants"
                    % (len(filesMaxReset), ",".join(unrecoverableStatus))
                )

            if filesWithDescendants:
                self.transLogger.warn(
                    "\t!!!!!!!! Transformation has descendants for files that are not marked as processed !!!!!!!!"
                )
                self.transLogger.warn("\tFiles with descendants:", ",".join(filesWithDescendants))
                self.transLogger.info(f"\tUpdating {len(filesWithDescendants)} files to '{'Processed'}'")
                result = self.__updateFileStatus(transformation, filesWithDescendants, "Processed")
                if not result["OK"]:
                    self.transLogger.error("\tRecovered files were not updated", result["Message"])

        return S_OK()

    #############################################################################
    def __getEligibleTransformations(self, status, typeList):
        """Select transformations of given status and type."""
        res = self.transClient.getTransformations(condDict={"Status": status, "Type": typeList})
        if not res["OK"]:
            return res
        transformations = {str(prod["TransformationID"]): prod["Type"] for prod in res["Value"]}
        return S_OK(transformations)

    #############################################################################
    def __selectTransformationFiles(self, transformation, statusList):
        """Select files, production jobIDs in specified file status for a given
        transformation."""
        # Until a query for files with timestamp can be obtained must rely on the
        # WMS job last update
        res = self.transClient.getTransformationFiles(
            condDict={"TransformationID": transformation, "Status": statusList}
        )
        if not res["OK"]:
            return res
        resDict = {}
        mandatoryKeys = {"LFN", "TaskID", "LastUpdate"}
        for fileDict in res["Value"]:
            missingKeys = mandatoryKeys - set(fileDict)
            if missingKeys:
                for key in missingKeys:
                    self.transLogger.warn(f"{key} is mandatory, but missing for:\n\t{str(fileDict)}")
            else:
                resDict[fileDict["LFN"]] = (fileDict["TaskID"], fileDict["Status"])
        return S_OK(resDict)

    #############################################################################
    def __obtainWMSJobIDs(self, transformation, fileDict, selectDelay, wmsStatusList):
        """Group files by the corresponding WMS jobIDs, check the corresponding
        jobs have not been updated for the delay time.

        Can't get into any mess because we start from files only in MaxReset
        / Assigned and check corresponding jobs.  Mixtures of files for jobs
        in MaxReset and Assigned statuses only possibly include some files
        in Unused status (not Processed for example) that will not be
        touched.
        """
        taskIDList = sorted({taskID for taskID, _status in fileDict.values()})
        self.transLogger.verbose(
            "The following %d task IDs correspond to the selected files:\n%s"
            % (len(taskIDList), ", ".join(str(taskID) for taskID in taskIDList))
        )

        jobFileDict = {}
        olderThan = datetime.datetime.utcnow() - datetime.timedelta(hours=selectDelay)

        res = self.transClient.getTransformationTasks(
            condDict={"TransformationID": transformation, "TaskID": taskIDList},
            older=olderThan,
            timeStamp="LastUpdateTime",
        )
        if not res["OK"]:
            self.transLogger.error("getTransformationTasks returned an error", f"{res['Message']}")
            return res

        mandatoryKeys = {"TaskID", "ExternalID", "LastUpdateTime", "ExternalStatus"}
        for taskDict in res["Value"]:
            missingKey = mandatoryKeys - set(taskDict)
            if missingKey:
                for key in missingKey:
                    self.transLogger.warn(f"Missing key {key} for job dictionary:\n\t{str(taskDict)}")
                continue

            taskID = taskDict["TaskID"]
            wmsID = taskDict["ExternalID"]
            wmsStatus = taskDict["ExternalStatus"]

            if not int(wmsID):
                self.transLogger.verbose(
                    f"TaskID {taskID}: status is {wmsStatus} (jobID = {wmsID}) so will not recheck with WMS"
                )
                continue

            # Exclude jobs not having appropriate WMS status - have to trust that production management status is correct
            if wmsStatus not in wmsStatusList:
                self.transLogger.verbose(
                    "Job %s is in status %s, not in %s so will be ignored"
                    % (wmsID, wmsStatus, ", ".join(wmsStatusList))
                )
                continue

            # Must map unique files -> jobs in expected state
            jobFileDict[wmsID] = [lfn for lfn, (tID, _st) in fileDict.items() if int(tID) == int(taskID)]

            self.transLogger.info(
                "Found %d files for taskID %s, jobID %s (%s), last update %s"
                % (len(jobFileDict[wmsID]), taskID, wmsID, wmsStatus, taskDict["LastUpdateTime"])
            )

        return S_OK(jobFileDict)

    #############################################################################

    def __removePendingRequestsJobs(self, jobFileDict):
        """Before doing anything check that no outstanding requests are pending for
        the set of WMS jobIDs."""
        jobs = list(jobFileDict)

        level = self.reqClient.log.getLevel()
        self.reqClient.log.setLevel("ERROR")
        result = self.reqClient.getRequestIDsForJobs(jobs)
        self.reqClient.log.setLevel(level)
        if not result["OK"]:
            return result

        if not result["Value"]["Successful"]:
            self.transLogger.verbose("None of the jobs have pending requests")
            return S_OK()

        for jobID, requestID in result["Value"]["Successful"].items():
            res = self.reqClient.getRequestStatus(requestID)
            if not res["OK"]:
                self.transLogger.error("Failed to get Status for Request", f"{requestID}:{res['Message']}")
            elif res["Value"] != "Done":
                # If we fail to get the Status or it is not Done, we must wait, so remove the job from the list.
                del jobFileDict[str(jobID)]
                self.transLogger.verbose(f"Removing jobID {jobID} from consideration until requests are completed")

        return S_OK()

    #############################################################################
    def __checkdescendants(self, transformation, jobFileDict):
        """Check BK descendants for input files, prepare list of actions to be
        taken for recovery."""

        jobsThatDidntProduceOutputs = []
        jobsThatProducedOutputs = []

        self.consChecks.prod = transformation
        for job, fileList in jobFileDict.items():
            result = self.consChecks.getDescendants(fileList)
            filesWithDesc = result[0]
            filesWithMultipleDesc = result[2]
            if filesWithDesc or filesWithMultipleDesc:
                jobsThatProducedOutputs.append(job)
            else:
                jobsThatDidntProduceOutputs.append(job)

        return jobsThatDidntProduceOutputs, jobsThatProducedOutputs

    ############################################################################
    def __updateFileStatus(self, transformation, fileList, fileStatus):
        """Update file list to specified status."""
        if not self.enableFlag:
            self.transLogger.info(
                "\tEnable flag is False, would have updated %d files to '%s' status for %s"
                % (len(fileList), fileStatus, transformation)
            )
            return S_OK()

        return self.transClient.setFileStatusForTransformation(int(transformation), fileStatus, fileList, force=False)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_RSS_Policy_TransferQualityPolicy."""
import unittest

import LHCbDIRAC.ResourceStatusSystem.Policy.TransferQualityPolicy as moduleTested

################################################################################


class TransferQualityPolicy_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.TransferQualityPolicy

    def tearDown(self):
        """Tear down."""

        del self.moduleTested
        del self.testClass


################################################################################


class TransferQualityPolicy_Success(TransferQualityPolicy_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""

        module = self.testClass()
        self.assertEqual("TransferQualityPolicy", module.__class__.__name__)

    def test_evaluate(self):
        """tests the method _evaluate."""

        module = self.testClass()

        res = module._evaluate({"OK": False, "Message": "Bo!"})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual("Bo!", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": None})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": []})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"A": 1}})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual('Missing "Name" key', res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1"}})
        self.assertTrue(res["OK"])
        self.assertEqual("Error", res["Value"]["Status"])
        self.assertEqual('Missing "Mean" key', res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1", "Mean": None}})
        self.assertTrue(res["OK"])
        self.assertEqual("Unknown", res["Value"]["Status"])
        self.assertEqual("No values to take a decision", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1", "Mean": 0}})
        self.assertTrue(res["OK"])
        self.assertEqual("Degraded", res["Value"]["Status"])
        self.assertEqual("TransferQuality: 0 -> Low", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1", "Mean": 70}})
        self.assertTrue(res["OK"])
        self.assertEqual("Degraded", res["Value"]["Status"])
        self.assertEqual("TransferQuality: 70 -> Mean", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1", "Mean": 95}})
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("TransferQuality: 95 -> High", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1failover", "Mean": 0}})
        self.assertTrue(res["OK"])
        self.assertEqual("Degraded", res["Value"]["Status"])
        self.assertEqual("TransferQuality: 0 -> Low", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1failover", "Mean": 70}})
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("TransferQuality: 70 -> Mean", res["Value"]["Reason"])

        res = module._evaluate({"OK": True, "Value": {"Name": "1failover", "Mean": 95}})
        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])
        self.assertEqual("TransferQuality: 95 -> High", res["Value"]["Reason"])


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TransferQualityPolicy_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TransferQualityPolicy_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

################################################################################



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
from copy import deepcopy

import pytest

import DIRAC
from DIRAC.Core.Base.AgentModule import AgentModule
from LHCbDIRAC.ProductionManagementSystem.Agent import ProductionStatusAgent as PSAModule
from LHCbDIRAC.ProductionManagementSystem.Agent.ProductionStatusAgent import ProductionStatusAgent


@pytest.fixture()
def psa(monkeypatch):
    monkeypatch.setattr(AgentModule, "__init__", lambda *a, **kw: None)
    gDoRealUpdate = PSAModule.gDoRealUpdate
    gDoRealTracking = PSAModule.gDoRealTracking
    PSAModule.gDoRealUpdate = False
    PSAModule.gDoRealTracking = False
    psa = ProductionStatusAgent(
        "ProductionManagement/ProductionStatusAgent",
        "ProductionManagement/ProductionStatusAgent",
    )
    psa.log = DIRAC.gLogger
    yield psa
    PSAModule.gDoRealUpdate = gDoRealUpdate
    PSAModule.gDoRealTracking = gDoRealTracking


anaProdBase = {
    -86392: {
        "type": "AnalysisProduction",
        "bkTotal": 900,
        "isDone": False,
        "isFinished": False,
        "master": 0,
        "prTotal": 475522154,
        "prods": {
            -139165: {
                "Events": 0,
                "Used": 0,
                "filesMaxReset": 0,
                "filesNotProcessed": 0,
                "filesProcessed": 12387,
                "filesTotal": 12387,
                "filesUnused": 0,
                "hasActiveInput": False,
                "inputIDs": [-75559, -75557, -79436, -77434],
                "isIdle": "Yes",
                "isProcIdle": "Yes",
                "isSimulation": False,
                "state": "Active",
            },
            -139166: {
                "Events": 900,
                "Used": 1,
                "filesMaxReset": 0,
                "filesNotProcessed": 0,
                "filesProcessed": 900,
                "filesTotal": 1275,
                "filesUnused": 375,
                "hasActiveInput": True,
                "inputIDs": [-139165],
                "isIdle": "No",
                "isProcIdle": "No",
                "isSimulation": False,
                "state": "Active",
            },
        },
    }
}

statesTestCases = []
# Do nothing
statesTestCases += [
    (
        {
            -85397: {
                "type": "Simulation",
                "master": -85318,
                "bkTotal": 249822,
                "prTotal": 250000,
                "isDone": False,
                "prods": {
                    -138943: {
                        "Events": 260179,
                        "Used": 0,
                        "state": "Active",
                        "isIdle": "Yes",
                        "isProcIdle": "Yes",
                        "isSimulation": True,
                        "filesTotal": 0,
                        "filesProcessed": 0,
                        "filesUnused": 0,
                        "filesMaxReset": 0,
                        "filesNotProcessed": 0,
                    },
                    -138944: {
                        "Events": 249822,
                        "Used": 1,
                        "state": "Active",
                        "isIdle": "No",
                        "isProcIdle": "No",
                        "isSimulation": False,
                        "filesTotal": 1177,
                        "filesProcessed": 1132,
                        "filesUnused": 45,
                        "filesMaxReset": 0,
                        "filesNotProcessed": 0,
                    },
                },
                "isFinished": False,
            }
        },
        {},
        [],
    ),
]

# --------------------
# Analysis Productions
# --------------------
# 1a. Active with all files processed goes to Idle
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139165: {"from": "Active", "to": "Idle"}},
        [],
    )
]
# 1b. Active with all files processed except those in MaxReset goes to Idle
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139165: {"from": "Active", "to": "Idle"}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["filesMaxReset"] = 1
statesTestCases[-1][0][-86392]["prods"][-139165]["filesProcessed"] -= 1

# 2a. Idle with all files processed goes to Completed
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139165: {"from": "Idle", "to": "Completed"}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
# 2b. Idle with some files in MaxReset does nothing
statesTestCases += [(deepcopy(anaProdBase), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139165]["filesMaxReset"] = 1
statesTestCases[-1][0][-86392]["prods"][-139165]["filesProcessed"] -= 1

# 3a. When the first transformation is done the second will flushed if all files are proccessed
statesTestCases += [
    (
        deepcopy(anaProdBase),
        {-139166: {"from": "Active", "to": "Flush"}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Finished"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = False
# 3b. It will even be flushed if some files are in NotProcessed
statesTestCases += [statesTestCases[-1]]
statesTestCases[-1][0][-86392]["prods"][-139166]["filesNotProcessed"] = 1
statesTestCases[-1][0][-86392]["prods"][-139166]["filesProcessed"] -= 1
# 3c. When the first transformation is done the second will not flushed if some files are still processing
statesTestCases += [(deepcopy(anaProdBase), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Finished"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = False
statesTestCases[-1][0][-86392]["prods"][-139166]["filesUnused"] -= 1
# 3d. When the first transformation is done the second will not flushed if some files are in maxReset
statesTestCases += [(deepcopy(anaProdBase), {-139166: {"from": "Active", "to": "Flush"}}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Finished"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = False
statesTestCases[-1][0][-86392]["prods"][-139166]["filesMaxReset"] = 1
statesTestCases[-1][0][-86392]["prods"][-139166]["filesUnused"] -= 1

anaProdBaseFinished = deepcopy(anaProdBase)
anaProdBaseFinished[-86392]["prods"][-139165]["state"] = "Finished"
anaProdBaseFinished[-86392]["prods"][-139166]["hasActiveInput"] = False
anaProdBaseFinished[-86392]["prods"][-139166]["isIdle"] = "Yes"
anaProdBaseFinished[-86392]["prods"][-139166]["filesUnused"] = 0
filesTotal = anaProdBaseFinished[-86392]["prods"][-139166]["filesTotal"]
anaProdBaseFinished[-86392]["prods"][-139166]["filesProcessed"] = filesTotal
# 4a. When both second transformation has processed all files it moves to Idle
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {"from": "Active", "to": "Idle"}},
        [],
    )
]
# 4b. It then moves to ValidatingOutput
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {"from": "Idle", "to": "ValidatingOutput"}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "Idle"
# 4c. It also moves to ValidatingOutput if there are files in NotProcessed
statesTestCases += [statesTestCases[-1]]
statesTestCases[-1][0][-86392]["prods"][-139166]["filesNotProcessed"] = 1
statesTestCases[-1][0][-86392]["prods"][-139166]["filesProcessed"] -= 1
# 4d. It then stays in ValidatingOutput until another agent handles it
statesTestCases += [(deepcopy(anaProdBaseFinished), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "ValidatingOutput"
# 4e. After validation it then moves from ValidatedOutput to Completed
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {"from": "ValidatedOutput", "to": "Completed"}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "ValidatedOutput"
# 4c. The production request is then marked as Done
statesTestCases += [(deepcopy(anaProdBaseFinished), {}, [-86392])]
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "Finished"
# 4d. When both second transformation has processed all files it moves to Idle if the input is still Active
statesTestCases += [
    (
        deepcopy(anaProdBaseFinished),
        {-139166: {"from": "Active", "to": "Idle"}},
        [],
    )
]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139165]["hasActiveInput"] = True
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = True
# 4e. When both second transformation has processed all files it stays in Idle if the input is still Active
statesTestCases += [(deepcopy(anaProdBaseFinished), {}, [])]
statesTestCases[-1][0][-86392]["prods"][-139165]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139165]["hasActiveInput"] = True
statesTestCases[-1][0][-86392]["prods"][-139166]["state"] = "Idle"
statesTestCases[-1][0][-86392]["prods"][-139166]["hasActiveInput"] = True


@pytest.mark.parametrize("prSummary, expected_updatedT, expected_updatedPr", statesTestCases)
def test_stateTransitions(psa, prSummary, expected_updatedT, expected_updatedPr):
    psa.prSummary = prSummary
    updatedT, updatedPr = {}, []
    psa._applyProductionRequestsLogic(updatedT, updatedPr)
    assert updatedT == expected_updatedT
    assert updatedPr == expected_updatedPr



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test class for the MCSimulationTestingAgent."""
import datetime
import importlib
import unittest
from unittest.mock import MagicMock, patch

from LHCbDIRAC.TransformationSystem.Agent.MCSimulationTestingAgent import MCSimulationTestingAgent
from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production
from DIRAC.Core.Workflow.Workflow import fromXMLString
from DIRAC import gLogger

try:
    with open("./src/LHCbDIRAC/TransformationSystem/Agent/test/testWF.xml") as fd:
        storedJobDescription = fd.read()
except OSError:
    with open("./TransformationSystem/Agent/test/testWF.xml") as fd:
        storedJobDescription = fd.read()


class MCSimulationTestingAgentTestCase(unittest.TestCase):
    def setUp(self):
        self.mockAM = MagicMock()
        self.agent = importlib.import_module("LHCbDIRAC.TransformationSystem.Agent.MCSimulationTestingAgent")
        self.agent.AgentModule = self.mockAM
        self.agent = MCSimulationTestingAgent()
        self.agent.log = gLogger
        self.agent.log.setLevel("DEBUG")
        self.transID = 1
        self.tasks = [
            {
                "TargetSE": "Unknown",
                "TransformationID": 1,
                "LastUpdateTime": datetime.datetime(2014, 7, 29, 12, 12, 13),
                "RunNumber": 0,
                "CreationTime": datetime.datetime(2014, 7, 29, 12, 12, 13),
                "ExternalID": "0",
                "ExternalStatus": "Running",
                "TaskID": 1,
            },
            {
                "TargetSE": "Unknown",
                "TransformationID": 1,
                "LastUpdateTime": datetime.datetime(2014, 7, 29, 12, 12, 13),
                "RunNumber": 0,
                "CreationTime": datetime.datetime(2014, 7, 29, 12, 12, 13),
                "ExternalID": "0",
                "ExternalStatus": "Created",
                "TaskID": 2,
            },
        ]
        self.report = {
            "subject": "MCSimulation Test Failure Report. TransformationID: " + str(self.transID),
            "body": [
                "MCSimulation Test Failure Report. TransformationID: " + str(self.transID),
                "",
                "Transformation:",
                "----------------------------------------------------------------------",
                "TransformationID: " + str(self.transID),
                "TransformationName: transName",
                "LastUpdate: 29/07/2014 13:06",
                "Status: New",
                "Description: description",
                "TransformationFamily: 0",
                "Plugin: Standard",
                "Type: MCSimulation",
                "AgentType: Manual",
                "GroupSize: 1",
                "MaxNumberOfTasks: 0",
                "AuthorDN: /DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sbidwell/CN=758039/CN=Simon Bidwell",
                "TransformationGroup: General",
                "InheritedFrom: 0",
                "CreationDate: 29/07/2014 13:06",
                "FileMask: ",
                "EventsPerTask: 0",
                "AuthorGroup: devGroup",
                "",
                "Number of Tasks: " + str(len(self.tasks)),
                "Tasks:",
                "----------------------------------------------------------------------",
                "TaskID: 1",
                "TargetSE: Unknown",
                "LastUpdateTime: 29/07/2014 12:12",
                "RunNumber: 0",
                "CreationTime: 29/07/2014 12:12",
                "ExternalID: 0",
                "ExternalStatus: Running",
                "",
                "TaskID: 2",
                "TargetSE: Unknown",
                "LastUpdateTime: 29/07/2014 12:12",
                "RunNumber: 0",
                "CreationTime: 29/07/2014 12:12",
                "ExternalID: 0",
                "ExternalStatus: Created",
                "",
            ],
        }

        self.transClientMock = MagicMock()
        self.transClientMock.getTransformations.return_value = {
            "OK": True,
            "Value": [
                {
                    "Body": "",
                    "LastUpdate": datetime.datetime(2014, 7, 29, 13, 6, 8),
                    "Status": "New",
                    "TransformationID": 1,
                    "Description": "description",
                    "TransformationFamily": 0,
                    "Plugin": "Standard",
                    "Type": "MCSimulation",
                    "AgentType": "Manual",
                    "GroupSize": 1,
                    "LongDescription": "longDescription",
                    "MaxNumberOfTasks": 0,
                    "Hot": 0,
                    "AuthorDN": "/DC=ch/DC=cern/OU=Organic Units/OU=Users/CN=sbidwell/CN=758039/CN=Simon Bidwell",
                    "TransformationName": "transName",
                    "TransformationGroup": "General",
                    "InheritedFrom": 0,
                    "CreationDate": datetime.datetime(2014, 7, 29, 13, 6, 8),
                    "FileMask": "",
                    "EventsPerTask": 0,
                    "AuthorGroup": "devGroup",
                }
            ],
        }
        self.transClientMock.getStoredJobDescription.return_value = {"OK": True, "Value": ((1, storedJobDescription),)}
        self.transClientMock.setTransformationParameter.return_value = {"OK": True}
        self.transClientMock.extendTransformation.return_value = {"OK": True, "Value": [3, 4, 5]}

        self.bkClientMock = MagicMock()
        self.bkClientMock.bulkJobInfo.return_value = {
            "OK": True,
            "Value": {
                "Failed": [],
                "Successful": {
                    "123": [
                        {
                            "ApplicationName": "Gauss",
                            "ApplicationVersion": "v30r14",
                            "CPUTIME": 1000.0,
                            "DIRACJobId": 1563495,
                            "DIRACVersion": "v2r13 build 3",
                            "EventInputStat": None,
                            "ExecTime": 2000.0,
                            "FirstEventNumber": 1,
                            "JobId": 123456,
                            "Location": "LCG.Glasgow.uk",
                            "Name": "00001764_00000195_5",
                            "NumberOfEvents": 100,
                            "Production": 1764,
                            "StatisticsRequested": None,
                            "TotalLumonosity": 0,
                            "WNCACHE": "1024KB",
                            "WNCPUHS06": 10.0,
                            "WNCPUPOWER": None,
                            "WNMEMORY": "8195868kB",
                            "WNMODEL": "DualCoreAMDOpteron(tm)",
                            "WORKERNODE": "node046.",
                        },
                        {
                            "ApplicationName": "Boole",
                            "ApplicationVersion": "v30r14",
                            "CPUTIME": 100.0,
                            "DIRACJobId": 1563495,
                            "DIRACVersion": "v2r13 build 3",
                            "EventInputStat": None,
                            "ExecTime": 200.0,
                            "FirstEventNumber": 1,
                            "JobId": 123456,
                            "Location": "LCG.Glasgow.uk",
                            "Name": "00001764_00000195_5",
                            "NumberOfEvents": 100,
                            "Production": 1764,
                            "StatisticsRequested": None,
                            "TotalLumonosity": 0,
                            "WNCACHE": "1024KB",
                            "WNCPUHS06": 10.0,
                            "WNCPUPOWER": None,
                            "WNMEMORY": "8195868kB",
                            "WNMODEL": "DualCoreAMDOpteron(tm)",
                            "WORKERNODE": "node046.",
                        },
                        {
                            "ApplicationName": "Bof",
                            "ApplicationVersion": "v30r14",
                            "CPUTIME": 100.0,
                            "DIRACJobId": 1563495,
                            "DIRACVersion": "v2r13 build 3",
                            "EventInputStat": None,
                            "ExecTime": 200.0,
                            "FirstEventNumber": 1,
                            "JobId": 123456,
                            "Location": "LCG.Glasgow.uk",
                            "Name": "00001764_00000195_5",
                            "NumberOfEvents": 100,
                            "Production": 1764,
                            "StatisticsRequested": None,
                            "TotalLumonosity": 0,
                            "WNCACHE": "1024KB",
                            "WNCPUHS06": 10.0,
                            "WNCPUPOWER": None,
                            "WNMEMORY": "8195868kB",
                            "WNMODEL": "DualCoreAMDOpteron(tm)",
                            "WORKERNODE": "node046.",
                        },
                        {
                            "ApplicationName": "Moore",
                            "ApplicationVersion": "v30r14",
                            "CPUTIME": 200.0,
                            "DIRACJobId": 1563495,
                            "DIRACVersion": "v2r13 build 3",
                            "EventInputStat": None,
                            "ExecTime": 200.0,
                            "FirstEventNumber": 1,
                            "JobId": 123456,
                            "Location": "LCG.Glasgow.uk",
                            "Name": "00001764_00000195_5",
                            "NumberOfEvents": 100,
                            "Production": 1764,
                            "StatisticsRequested": None,
                            "TotalLumonosity": 0,
                            "WNCACHE": "1024KB",
                            "WNCPUHS06": 10.0,
                            "WNCPUPOWER": None,
                            "WNMEMORY": "8195868kB",
                            "WNMODEL": "DualCoreAMDOpteron(tm)",
                            "WORKERNODE": "node046.",
                        },
                    ]
                },
            },
            "rpcStub": (
                (
                    "Bookkeeping/BookkeepingManager",
                    {
                        "delegatedDN": "/DC=ch/CN=Simon Bidwell",
                        "delegatedGroup": "lhcb_user",
                        "keepAliveLapse": 150,
                        "skipCACheck": False,
                        "timeout": 3600,
                    },
                ),
                "bulkJobInfo",
                ([123, 456],),
            ),
        }

        self.notifyClientMock = MagicMock()
        self.notifyClientMock.sendMail.return_value = {"OK": True, "Value": "The mail was succesfully sent"}

        self.operationsMock = MagicMock()
        self.operationsMock.getValue.return_value = "sbidwell"

        self.agent.transClient = self.transClientMock
        self.agent.bkClient = self.bkClientMock
        self.agent.notifyClient = self.notifyClientMock
        self.agent.operations = self.operationsMock
        self.agent.email = "myEmail@cern.ch"

    def tearDown(self):
        pass

    def test_send_report(self):
        res = self.agent._sendReport(self.report)
        self.assertEqual(res, None)

    def test_update_workflow(self):
        CPUe = 1
        MCCpu = 25
        s1 = """<Parameter name="CPUe" type="string" linked_module="" linked_parameter="" in="True" out="False" """
        s1 += """description="CPU time per event"><value><![CDATA[1]]></value></Parameter>\n"""
        CPUe_xml = self.test_workflow = s1
        s2 = """<Parameter name="maxNumberOfEvents" type="string" linked_module="" linked_parameter="" """
        s2 += """in="True" out="False" description="Maximum number of events to produce (Gauss)">"""
        s2 += """<value><![CDATA[150000]]></value></Parameter>\n"""
        max_e_xml = self.test_workflow = s2
        res = self.agent._updateWorkflow(self.transID, CPUe, MCCpu)
        self.assertTrue(res["OK"])
        prod = Production()
        prod.LHCbJob.workflow = fromXMLString(res["Value"])
        cpue_param = prod.LHCbJob.workflow.findParameter("CPUe")
        max_e_param = prod.LHCbJob.workflow.findParameter("maxNumberOfEvents")
        self.assertEqual(CPUe_xml, cpue_param.toXML())
        self.assertEqual(max_e_xml, max_e_param.toXML())

    @patch("LHCbDIRAC.TransformationSystem.Agent.MCSimulationTestingAgent.getEventsToProduce")
    def test_calculate_parameters(self, _patch_mock):
        #     expected_max_e = 100
        # mock to make getEventsToProduce to return 100
        #     patch_mock.return_value = expected_max_e
        res = self.agent._calculateParameters(self.tasks)
        self.assertTrue(res["OK"])
        self.assertEqual(res["Value"]["CPUe"], 130.0)


#     self.assertEqual( res['Value']['max_e'], expected_max_e )


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(MCSimulationTestingAgentTestCase)
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    Script.setUsageMessage(
        """
  Check the integrity of the state of the storages and information in the File Catalogs
  for a given directory or a collection of directories.

  Usage:
    %s  (<options>|<cfgFile>) <dir | fileContainingDirs> (mandatory)
  """
        % Script.scriptName
    )

    Script.parseCommandLine()

    import DIRAC
    from DIRAC import gLogger
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    from LHCbDIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient

    fc = FileCatalog()
    integrity = DataIntegrityClient()
    gLogger.setLevel("INFO")

    args = Script.getPositionalArgs()
    if len(args) < 1:
        print("Please provide a directory or a file containing directories")
        Script.showHelp(exitCode=1)
    inputNames = args

    directories = []
    for inputFileName in inputNames:
        try:
            with open(inputFileName) as inputFile:
                stringIn = inputFile.read()
                directories += stringIn.splitlines()
        except Exception:
            directories.append(inputFileName)

    ######################################################
    #
    # This check performs Catalog->BK and Catalog->SE for possible output directories
    #
    res = fc.exists(directories)
    if not res["OK"]:
        gLogger.error(res["Message"])
        DIRAC.exit(-2)
    for directory, error in res["Value"]["Failed"]:
        gLogger.error("Failed to determine existence of directory", f"{directory} {error}")
    if res["Value"]["Failed"]:
        DIRAC.exit(-2)
    directoryExists = res["Value"]["Successful"]
    for directory in sorted(directoryExists):
        if not directoryExists[directory]:
            continue
        gLogger.info(f"Checking the integrity of {directory}")
        iRes = integrity.catalogDirectoryToBK(directory)
        if not iRes["OK"]:
            gLogger.error("Error getting directory content:", iRes["Message"])
            continue
        catalogDirMetadata = iRes["Value"]["CatalogMetadata"]
        catalogDirReplicas = iRes["Value"]["CatalogReplicas"]
        catalogMetadata = {}
        catalogReplicas = {}
        for lfn in catalogDirMetadata:
            # if not lfn in bk2catalogMetadata:
            catalogMetadata[lfn] = catalogDirMetadata[lfn]
            if lfn in catalogDirReplicas:
                catalogReplicas[lfn] = catalogDirReplicas[lfn]
        if not catalogMetadata:
            continue
        res = integrity.checkPhysicalFiles(catalogReplicas, catalogMetadata, [])
        if not res["OK"]:
            gLogger.error("Error checking physical files:", res["Message"])
            continue


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""interpret the data taking conditions."""


SUBDETECTOR_SHORT_NAMES = {
    "ECAL": "EC",
    "HCAL": "HC",
    "MUON": "MU",
    "PLUME": "PL",
    "RICH1": "R1",
    "RICH2": "R2",
    "SCIFI": "SF",
    "SPD_PRS": "SP",
    "VELO": "VE",
}


# For some reason, there are runs that were registered
# with this energy or greater, and where the beam should be
# considered off in the description...
MAX_BEAM_ENERGY = 7864


def generateConditionDescription(dataTakingDict: dict[str, str], configName: str) -> str:
    """Based on the sub detector status and the configuration name (partition),
    generate a condition description string.

    Rules:

    * If the partition is LHCb:
      <BeamEnergy>-<VeloPosition>-<MagPolarity>(-Excl-<list of excluded subdetectors>)

    * else:
      <BeamEnergy>-<MagPolarity>

    """

    beamEnergyDesc = "BeamOff"

    try:
        beamEnergyInt = int(float(dataTakingDict.get("BeamEnergy", "0").strip()))
        if 0 < beamEnergyInt < MAX_BEAM_ENERGY:
            beamEnergyDesc = f"Beam{beamEnergyInt}GeV"
    # Raised if it is not a number
    except ValueError:
        pass

    magPolarityDesc = f"Mag{dataTakingDict['MagneticField'].capitalize()}"

    if configName == "LHCb":
        veloPositionDesc = f"Velo{dataTakingDict['VeloPosition'].capitalize()}"

        excludedSubDetectors = []
        # Check all the excluded sub detectors
        for subDetector, subDetectorStatus in dataTakingDict.items():
            # The L0 trigger and TDET can be included or not, but they are not
            # subdetector so skip it
            if subDetector in ("L0", "TDET", "PLUME"):
                continue
            if subDetectorStatus == "NOT INCLUDED":
                excludedSubDetectors.append(SUBDETECTOR_SHORT_NAMES.get(subDetector, subDetector))

        descriptionParameters = [beamEnergyDesc, veloPositionDesc, magPolarityDesc]
        if excludedSubDetectors:
            excludedSubDetectorDesc = "Excl-" + "-".join(sorted(excludedSubDetectors))
            descriptionParameters.append(excludedSubDetectorDesc)

    else:
        descriptionParameters = [beamEnergyDesc, magPolarityDesc]

    return "-".join(descriptionParameters)



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""UploadMC module is used to upload to ES the json files for MC statistics."""
import os
import json
from datetime import datetime, timezone

from DIRAC import S_OK, S_ERROR, gLogger

from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase
from LHCbDIRAC.ProductionManagementSystem.Client.MCStatsClient import MCStatsClient
from LHCbDIRAC.Core.Utilities.XMLSummaries import XMLSummary
from LHCbDIRAC.Core.Utilities.GeneratorLog import GeneratorLog


class UploadMC(ModuleBase):
    """Upload to LogSE."""

    def __init__(self):
        """Module initialization."""

        self.log = gLogger.getSubLogger("UploadMC")
        super().__init__(self.log)

    def _resolveInputVariables(self):
        """standard method for resolving the input variables."""

        super()._resolveInputVariables()

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main executon method.

        Takes care of indexing JSON files on ElasticSearch databases
        All these uploads are controlled by flags that can be found in
        'Operations/<setup>/Productions/'

        1) errors => /Operations/<setup>/Productions/UploadES_errors (default: True)
        2) Summary.xml -> JSON => /Operations/<setup>/Productions/UploadES_XMLSummary (default: False)
        3) GeneratorLog.xml -> JSON => /Operations/<setup>/Productions/UploadES_GeneratorLog (default: False)
        4) PRMon => /Operations/<setup>/Productions/UploadES_PrMon (default: False)

        """
        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            self._resolveInputVariables()

            mcStatsClient = MCStatsClient()

            # DBs list
            # db = {
            #     'XMLSummary': elasticApplicationSummaryDB,
            #     'errors': ElasticLogErrorsDB,
            #     'generatorLog': elasticMCGeneratorLogDB,
            #     'prMon': elasticMCGeneratorLogDB,
            # }

            # 1) MC errors
            # looking for json files that are 'self.jobID_Errors_Gauss.json'
            # TODO: it could be extended to at least Boole
            fn = f"{self.jobID}_Errors_Gauss.json"
            if os.path.exists(fn):
                with open(fn) as fd:
                    try:
                        jsonData = json.load(fd)
                        self.log.verbose("Content of JSON file", f"{fn}: {jsonData}")
                        errorList = []  # Fill this list to send the data in a bulk
                        for i, error in jsonData.items():
                            errorList.append(error)
                        if self._enableModule() and self.opsH.getValue("Productions/UploadES_errors", True):
                            res = mcStatsClient.set("errors", errorList)
                            if not res["OK"]:
                                self.log.error(
                                    "MC Error data not set, exiting without affecting workflow status",
                                    f"{str(errorList)}: {res['Message']}",
                                )
                        else:
                            # At this point we can see exactly what the module would have uploaded
                            self.log.info("Module disabled", f"would have attempted to upload the following file {fn}")
                    except Exception as ve:
                        self.log.error(repr(ve))
                        self.log.verbose(f"Exception loading the JSON file: content of {fn} follows")
                        self.log.verbose(fd.read)
                        # do not fail the job for this
                        # raise
            else:
                self.log.info("Gauss errors JSON file not found", fn)

            # 2) summary JSON files
            # looking for xml files that are 'summaryGauss_self.production_id_self.prod_job_id_1.xml'
            xmlfl = f"summaryGauss_{self.production_id}_{self.prod_job_id}_1.xml"
            if os.path.exists(xmlfl):
                try:
                    xmlData = XMLSummary(xmlfl)
                    xmlData.xmltojson()
                    # At this point 'summaryGauss_self.production_id_self.prod_job_id_1.json' should have been created
                    jsonfl = f"summaryGauss_{self.production_id}_{self.prod_job_id}_1.json"
                    with open(jsonfl) as JS:
                        jsonData = json.load(JS)
                        ids = dict()
                        ids["JobID"] = self.jobID
                        ids["ProductionID"] = self.production_id
                        ids["prod_job_id"] = self.prod_job_id
                        jsonData["Counters"]["ID"] = ids
                        with open(jsonfl, "w", encoding="utf-8") as output:
                            output.write(str(json.dumps(jsonData, indent=2)))

                        self.log.verbose("Content of JSON file", f"{jsonfl}: {jsonData}")
                        if self._enableModule() and self.opsH.getValue("Productions/UploadES_XMLSummary", False):
                            res = mcStatsClient.set("XMLSummary", jsonData)
                            if not res["OK"]:
                                self.log.error(
                                    "Gauss Summaries data not set, exiting without affecting workflow status",
                                    f"{str(jsonData)}: {res['Message']}",
                                )
                        else:
                            # At this point we can see exactly what the module would have uploaded
                            self.log.info(
                                "Module disabled", f"would have attempted to upload the following file {jsonfl}"
                            )

                except Exception:
                    self.log.exception("Exception creating/loading the XMLSummary JSON file")
                    # do not fail the job for this

            else:
                self.log.info("XML Gauss summary file not found", xmlfl)

            # 3) Generator logs
            # looking for xml files that are 'GeneratorLog.xml'
            xmlfile = "GeneratorLog.xml"
            if os.path.exists(xmlfile):
                try:
                    jsonfile = f"GeneratorLog_{self.production_id}_{self.prod_job_id}.json"
                    xmlData = GeneratorLog()
                    xmlData.generatorLogJson(jsonfile)
                    # At this point 'GeneratorLog_self.production_id_self.prod_job_id.json' should have been created
                    with open(jsonfile) as JS:
                        jsonData = json.load(JS)
                        ids = dict()
                        ids["JobID"] = self.jobID
                        ids["ProductionID"] = self.production_id
                        ids["prod_job_id"] = self.prod_job_id
                        jsonData["generatorCounters"]["ID"] = ids
                        jsonData["generatorCounters"]["timestamp"] = str(datetime.now(tz=timezone.utc))
                        with open(jsonfile, "w", encoding="utf-8") as output:
                            output.write(str(json.dumps(jsonData)))

                        self.log.verbose("Content of JSON file", f"{jsonfile}: {jsonData}")
                        if self._enableModule() and self.opsH.getValue("Productions/UploadES_GeneratorLog", False):
                            res = mcStatsClient.set("generatorLog", jsonData)
                            if not res["OK"]:
                                self.log.error(
                                    "Generator Log data not set, exiting without affecting workflow status",
                                    f"{str(jsonData)}: {res['Message']}",
                                )
                        else:
                            # At this point we can see exactly what the module would have uploaded
                            self.log.info(
                                "Module disabled", f"would have attempted to upload the following file {jsonfile}"
                            )

                except Exception:
                    self.log.exception("Exception creating/loading the GeneratorLog JSON file")
                    # do not fail the job for this

            else:
                self.log.info("XML GeneratorLog file not found", xmlfile)

            # 4) PRmon JSON files (only for Gauss)
            # looking for prmon files that are 'prmon_Gauss.json'
            prmonFile = "prmon_Gauss.json"
            if os.path.exists(prmonFile):
                with open(prmonFile) as JS:
                    try:
                        jsonData = json.load(JS)
                        jsonData["JobID"] = self.jobID
                        jsonData["ProductionID"] = self.production_id
                        jsonData["prod_job_id"] = self.prod_job_id
                        self.log.verbose("Content of JSON file", f"{prmonFile}: {jsonData}")
                        if self._enableModule() and self.opsH.getValue("Productions/UploadES_PrMon", False):
                            res = mcStatsClient.set("prMon", jsonData)
                            if not res["OK"]:
                                self.log.error(
                                    "prmon Metrics data not set, exiting without affecting workflow status",
                                    f"{str(jsonData)}: {res['Message']}",
                                )
                        else:
                            # At this point we can see exactly what the module would have uploaded
                            self.log.info(
                                "Module disabled", f"would have attempted to upload the following file {prmonFile}"
                            )
                    except Exception as ve:
                        self.log.error(repr(ve))
                        self.log.verbose(f"Exception loading the JSON file: content of {prmonFile} follows")
                        self.log.verbose(JS.read())
                        # do not fail the job for this
                        # raise
            else:
                self.log.info("prmon JSON file not found", prmonFile)

            return S_OK()

        except Exception as e:
            self.log.exception("Failure in UploadMC execute module", lException=e)
            return S_ERROR(repr(e))

        finally:
            super().finalize()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import DIRAC
from DIRAC.Core.Base.Script import Script


def usage():
    """usage.

    Prints script usage
    """
    print(f"Usage: {Script.scriptName} <LFN> [<LFN>] [--ProductionID=<ID>] [Try -h,--help for more information]")
    DIRAC.exit(2)


@Script()
def main():
    Script.registerSwitch(
        "",
        "ProductionID=",
        "Restrict query to given production ID (default is to show status for all)",
    )
    Script.parseCommandLine(ignoreErrors=True)

    from LHCbDIRAC.Interfaces.API.DiracProduction import DiracProduction

    prodID = ""
    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() == "productionid":
            prodID = switch[1]

    args = Script.getPositionalArgs()

    if len(args) < 1:
        usage()

    if prodID:
        try:
            prodID = int(prodID)
        except Exception as x:
            print("ERROR ProductionID should be an integer")
            DIRAC.exit(2)

    diracProd = DiracProduction()
    exitCode = 0
    result = diracProd.checkFilesStatus(args, prodID, printOutput=False)
    if not result["OK"]:
        print(f"ERROR {result['Message']}")
        exitCode = 2

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Just couple utilities."""
import json
import os
import sqlite3

import DIRAC
from DIRAC import gConfig, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getUserOption, getUsersInGroup
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.ConfigurationSystem.Client import PathFinder


def _getMemberMails(group):
    """get members mails."""
    members = getUsersInGroup(group)
    if members:
        emails = []
        for user in members:
            email = getUserOption(user, "Email")
            if email:
                emails.append(email)
        return emails


def _aggregate(reqId, reqType, reqWG, reqName, SimCondition, ProPath, groups, informPeoples):
    cacheFile = os.path.join(DIRAC.rootPath, "work/ProductionManagement/cache.db")

    with sqlite3.connect(cacheFile) as conn:
        try:
            conn.execute(
                """CREATE TABLE IF NOT EXISTS ProductionManagementCache(
                    reqId VARCHAR(64) NOT NULL,
                    reqType VARCHAR(64) NOT NULL,
                    reqWG VARCHAR(64) DEFAULT "",
                    reqName VARCHAR(64) NOT NULL,
                    SimCondition VARCHAR(64) DEFAULT "",
                    ProPath VARCHAR(64) DEFAULT "",
                    thegroup VARCHAR(64) DEFAULT "",
                    reqInform VARCHAR(64) DEFAULT ""
                   );"""
            )

        except sqlite3.OperationalError:
            gLogger.error("Email cache database is locked")

        for group in groups:
            conn.execute(
                "INSERT INTO ProductionManagementCache (reqId, reqType, reqWG, reqName,\
             SimCondition, ProPath, thegroup, reqInform)"
                " VALUES (?, ?, ?, ?, ?, ?, ?, ?)",
                (reqId, reqType, reqWG, reqName, SimCondition, ProPath, group, informPeoples),
            )

            conn.commit()


def informPeople(rec, oldstate, state, author, inform):
    """inform utility."""

    if not state or state == "New":
        return  # was no state change or resurrect

    reqId = rec["RequestID"]
    csS = PathFinder.getServiceSection("ProductionManagement", "ProductionRequest")
    if not csS:
        gLogger.error("No ProductionRequest section in configuration")
        return

    fromAddress = gConfig.getValue(f"{csS}/fromAddress", "")
    if not fromAddress:
        gLogger.error(f"No fromAddress is defined in CS path {csS}/fromAddress")
        return
    sendNotifications = gConfig.getValue(f"{csS}/sendNotifications", "Yes")
    if sendNotifications != "Yes":
        gLogger.info("No notifications will be sent")
        return

    footer = "\n\nNOTE: This is an automated notification."
    footer += " Please do not reply.\n"

    footer += f"You can find the production request {reqId} in the DIRAC Web portal:\n"
    footer += f"https://lhcb-portal-dirac.cern.ch/DIRAC/s:{PathFinder.getDIRACSetup()}/g:"

    ppath = "/?view=tabs&url_state=1|*LHCbDIRAC.ProductionRequestManager.classes.ProductionRequestManager:,\n\n"

    ppath += "The request details:\n"
    ppath += f"  Type: {str(rec['RequestType'])}\n"
    ppath += f"  Name: {str(rec['RequestName'])}\n"
    ppath += f"  Conditions: {str(rec['SimCondition'])}\n"
    ppath += f"  Processing pass: {str(rec['ProPath'])}\n"

    gLogger.info(f".... {ppath} ....")

    authorMail = getUserOption(author, "Email")
    if authorMail:
        if state not in ["BK Check", "Submitted"]:
            if state == "BK OK":
                subj = f"DIRAC: please sign your Production Request {reqId}"
                body = "\n".join(
                    [
                        "Customized Simulation Conditions in your request was registered. ",
                        "Since the Bookkeeping expert could make changes in your request, ",
                        "you are asked to confirm it.",
                    ]
                )
            else:
                subj = "DIRAC: the state of Production Request {} is changed to '{}'; {};{}".format(
                    reqId,
                    state,
                    rec.get("RequestWG", ""),
                    rec.get("RequestName", ""),
                )
                body = "\n".join(["The state of your request is changed. ", "This mail is for information only."])
            notification = NotificationClient()
            res = notification.sendMail(authorMail, subj, body + footer + "lhcb_user" + ppath, fromAddress, True)
            if not res["OK"]:
                gLogger.error(f"_inform_people: can't send email: {res['Message']}")

    if inform:
        subj = "DIRAC: the state of {} Production Request {} is changed to '{}'; {};{}".format(
            rec["RequestType"],
            reqId,
            state,
            rec.get("RequestWG", ""),
            rec.get("RequestName", ""),
        )
        body = "\n".join(["You have received this mail because you are " "in the subscription list for this request"])
        for x in inform.replace(" ", ",").split(","):
            if x:
                if x.find("@") > 0:
                    eMail = x
                else:
                    eMail = getUserOption(x, "Email")
                if eMail:
                    notification = NotificationClient()
                    res = notification.sendMail(eMail, subj, body + footer + "lhcb_user" + ppath, fromAddress, True)
                    if not res["OK"]:
                        gLogger.error(f"_inform_people: can't send email: {res['Message']}")

    if state == "Accepted":
        subj = "DIRAC: the Production Request {} is accepted; {};{}".format(
            reqId,
            rec.get("RequestWG", ""),
            rec.get("RequestName", ""),
        )
        body = "\n".join(
            ["The Production Request is signed and ready to process. ", "You are informed as member of %s group"]
        )
        groups = ["lhcb_prmgr"]

        for group in groups:
            for man in _getMemberMails(group):
                notification = NotificationClient()
                res = notification.sendMail(man, subj, body % group + footer + group + ppath, fromAddress, True)
                if not res["OK"]:
                    gLogger.error(f"_inform_people: can't send email: {res['Message']}")

    elif state == "PPG OK" and oldstate == "Accepted":
        subj = "DIRAC: returned Production Request {}; {};{}".format(
            reqId,
            rec.get("RequestWG", ""),
            rec.get("RequestName", ""),
        )
        body = "\n".join(
            [
                "Production Request is returned by Production Manager. ",
                "As member of %s group, you are asked to correct and sign ",
                "or to reject it.",
                "",
                "In case some other member of the group has already ",
                "done that, please ignore this mail.",
            ]
        )
        groups = ["lhcb_tech"]

        for group in groups:
            for man in _getMemberMails(group):
                notification = NotificationClient()
                res = notification.sendMail(man, subj, body % group + footer + group + ppath, fromAddress, True)
                if not res["OK"]:
                    gLogger.error(f"_inform_people: can't send email: {res['Message']}")

    elif state == "BK Check":
        groups = ["lhcb_bk"]

        _aggregate(
            reqId,
            rec.get("RequestType", ""),
            rec.get("RequestWG", ""),
            rec.get("RequestName", ""),
            rec["SimCondition"],
            rec["ProPath"],
            groups,
            rec.get("reqInform", inform),
        )

    elif state == "Submitted":
        groups = ["lhcb_ppg", "lhcb_tech"]
        _aggregate(
            reqId,
            rec.get("RequestType", ""),
            rec.get("RequestWG", ""),
            rec.get("RequestName", ""),
            rec["SimCondition"],
            rec["ProPath"],
            groups,
            rec.get("reqInform", inform),
        )

    else:
        return


def unpackOptionFile(optionsFile):
    if optionsFile.startswith("{"):
        return json.loads(optionsFile)
    else:
        return optionsFile.split(";")



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Use dirac-proxy-init to get a proxy."""
import os
import sys
from DIRAC.Core.Base.Script import Script


@Script()
def main():
    if os.getenv("X509_CERT_DIR") is None:
        sys.exit("the variable X509_CERT_DIR do not exist")

    if not os.path.isdir(os.environ["X509_CERT_DIR"]):
        sys.exit(f"the directory {os.environ['X509_CERT_DIR']} does not exist")

    if os.getenv("X509_VOMS_DIR") is None:
        sys.exit("the variable X509_VOMS_DIR do not exist")

    if not os.path.isdir(os.environ["X509_VOMS_DIR"]):
        sys.exit(f"the directory {os.environ['X509_VOMS_DIR']} does not exist")

    out = os.system("dirac-proxy-init -o LogLevel=NOTICE --strict '%s'" % "' '".join(sys.argv[1:]))
    sys.exit(int(out / 256))


if __name__ == "__main__":
    main()



#! /usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set the destination for a set of runs, based on the majority of reco
output."""
import DIRAC
from DIRAC.Core.Base.Script import Script


def execute(dmScript):
    """Parse the options and execute the script."""
    from DIRAC import gLogger
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import ProgressBar

    bkQuery = dmScript.getBKQuery()
    fileType = bkQuery.getFileTypeList()
    if not set(fileType) & {"FULL.DST", "RDST", "SDST"}:
        gLogger.error("Please provide a reconstruction BK path")
        DIRAC.exit(1)

    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
    from DIRAC.DataManagementSystem.Client.DataManager import DataManager
    from DIRAC.Core.Utilities.List import breakListIntoChunks
    from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
    from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, resolveSEGroup

    bk = BookkeepingClient()
    tr = TransformationClient()
    dm = DataManager()
    dmsHelper = DMSHelpers()

    bkQueryDict = bkQuery.getQueryDict()
    gLogger.notice("For BK Query:", str(bkQueryDict))
    progressBar = ProgressBar(1, title="Running BK query...", step=1)
    res = bk.getFilesWithMetadata(bkQueryDict)
    if not res["OK"]:
        gLogger.error("Error getting files from BK", res["Message"])
        DIRAC.exit(2)

    if "ParameterNames" not in res.get("Value", {}):
        gLogger.error("\nNo metadata found")
        DIRAC.exit(3)
    parameterNames = res["Value"]["ParameterNames"]
    info = res["Value"]["Records"]
    progressBar.endLoop(f"Obtained {len(info)} files")
    lfns = []
    runLFNs = {}
    for item in info:
        metadata = dict(zip(parameterNames, item))
        lfn = metadata["FileName"]
        lfns.append(lfn)
        runLFNs.setdefault(metadata["RunNumber"], []).append(lfn)

    chunkSize = 1000
    progressBar = ProgressBar(len(lfns), title=f"Getting replicas of {len(lfns)} files", chunk=chunkSize)
    replicas = {}
    errors = {}
    for lfnChunk in breakListIntoChunks(lfns, chunkSize):
        progressBar.loop()
        res = dm.getReplicas(lfnChunk, getUrl=False)
        if not res["OK"]:
            errors.setdefault(res["Message"], []).extend(lfnChunk)
        else:
            replicas.update(res["Value"]["Successful"])
            for lfn, error in res["Value"]["Failed"].items():
                errors.setdefault(error, []).append(lfn)
    progressBar.endLoop()
    for error, lfns in errors.items():
        gLogger.error(error, f"for {len(lfns)} files")

    tier1RDST = set(resolveSEGroup("Tier1-RDST"))
    setOK = 0
    errors = {}
    progressBar = ProgressBar(len(runLFNs), title=f"Defining destination for {len(runLFNs)} runs", step=10)
    for run, lfns in runLFNs.items():
        progressBar.loop()
        res = tr.getDestinationForRun(run)
        if res.get("Value"):
            errors.setdefault("Destination already set", []).append(str(run))
            continue
        # print 'Run', run, len( lfns ), 'Files', lfns[:3]
        seCounts = {}
        for lfn in lfns:
            for se in tier1RDST.intersection(replicas.get(lfn, [])):
                seCounts[se] = seCounts.setdefault(se, 0) + 1
        # print seCounts
        maxi = 0
        seMax = None
        for se, count in seCounts.items():
            if count > maxi:
                seMax = se
                maxi = count
        if not seMax:
            errors.setdefault("No SE found, use CERN-RDST", []).append(str(run))
            seMax = "CERN-RDST"
        # SE found, get its site
        res = dmsHelper.getLocalSiteForSE(seMax)
        if res["OK"]:
            site = res["Value"]
            res = tr.setDestinationForRun(run, site)
            if not res["OK"]:
                errors.setdefault(res["Message"], []).append(str(run))
            else:
                setOK += 1
    progressBar.endLoop("Successfully set destination for %d runs" % setOK)
    for error, runs in errors.items():
        gLogger.error(error, f"for runs {','.join(runs)}")


@Script()
def main():
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerBKSwitches()

    Script.setUsageMessage(
        __doc__
        + "\n".join(
            [
                "Usage:",
                f"  {Script.scriptName} [option|cfgfile]",
            ]
        )
    )

    Script.parseCommandLine(ignoreErrors=False)

    execute(dmScript)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Set files Removed in a transformation."""

from DIRAC.Core.Base.Script import Script
from DIRAC.TransformationSystem.Utilities.ScriptUtilities import getTransformations


@Script()
def main():
    import DIRAC
    from DIRAC import gLogger
    from LHCbDIRAC.DataManagementSystem.Client.DMScript import DMScript

    dmScript = DMScript()
    dmScript.registerFileSwitches()

    Script.parseCommandLine(ignoreErrors=True)

    Script.setUsageMessage(
        "\n".join(
            [
                __doc__,
                "Usage:",
                "Set files as Removed in a (set of) transformations",
                f"  {Script.scriptName} [option|cfgfile] ...",
            ]
        )
    )

    transList = getTransformations(Script.getPositionalArgs())
    if not transList:
        DIRAC.exit(1)

    requestedLFNs = dmScript.getOption("LFNs")
    if not requestedLFNs:
        gLogger.always("No files to add")
        DIRAC.exit(1)

    from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    trClient = TransformationClient()
    rc = 0
    for transID in transList:
        res = trClient.setFileStatusForTransformation(transID, "Removed", requestedLFNs, force=True)
        if res["OK"]:
            gLogger.always(
                "Successfully set %d files%s Removed in transformation %d"
                % (
                    len(res["Value"]),
                    (" (out of %d)" % len(requestedLFNs)) if len(res["Value"]) != len(requestedLFNs) else "",
                    transID,
                )
            )
        else:
            gLogger.always(
                "Failed to set %d files Removed in transformation %d" % (len(requestedLFNs), transID), res["Message"]
            )
            rc = 2
    DIRAC.exit(rc)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""The Storage History Agent will create a summary of the storage usage DB
grouped by processing pass or other interesting parameters.

"""
import os
import time
import copy
import datetime
from collections import defaultdict

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.File import mkDir, convertSizeUnits
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.AccountingSystem.Client.DataStoreClient import gDataStoreClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Core.Utilities.List import breakListIntoChunks

from LHCbDIRAC.AccountingSystem.Client.Types.UserStorage import UserStorage
from LHCbDIRAC.AccountingSystem.Client.Types.Storage import Storage
from LHCbDIRAC.AccountingSystem.Client.Types.DataStorage import DataStorage
from LHCbDIRAC.BookkeepingSystem.Client.BookkeepingClient import BookkeepingClient
from LHCbDIRAC.DataManagementSystem.Client.DataUsageClient import DataUsageClient
from LHCbDIRAC.DataManagementSystem.Client.StorageUsageClient import StorageUsageClient


def _standardDirectory(dirPath):
    """Add a "/" at the end of the directory name if not present"""
    return os.path.join(dirPath, "")


def _fillMetadata(dictToFill, metadataValue):
    """Fill the dictionary to send to the accounting.

    If metadataValue is a string then set all the values of dictToFill, to
    this value if metadataValue is a dictionary then set each value of
    dictToFill to the corresponding value of metadataValue
    """
    # ds = DataStorage()
    # keyList = ds.keyFieldsList
    # this is the list of attributes returned by the Bookkeeping for a given directory
    keyList = (
        "ConfigName",
        "ConfigVersion",
        "FileType",
        "Production",
        "ProcessingPass",
        "ConditionDescription",
        "EventType",
        "Visibility",
    )
    if isinstance(metadataValue, str):
        for k in keyList:
            dictToFill[k] = metadataValue
    elif isinstance(metadataValue, dict):
        for k in keyList:
            dictToFill[k] = metadataValue.get(k, "na")


class StorageHistoryAgent(AgentModule):
    def initialize(self):
        """Sets defaults."""
        self.am_setOption("PollingTime", 43200)
        if self.am_getOption("DirectDB", False):
            from LHCbDIRAC.DataManagementSystem.DB.StorageUsageDB import StorageUsageDB

            self.__stDB = StorageUsageDB()
        else:
            self.__stDB = StorageUsageClient()
        self.__workDirectory = self.am_getOption("WorkDirectory")
        mkDir(self.__workDirectory)
        self.log.info(f"Working directory is {self.__workDirectory}")

        self.__ignoreDirsList = self.am_getOption("Ignore", [])
        self.log.info(f"List of directories to ignore for the DataStorage accounting: {self.__ignoreDirsList} ")

        self.__bkClient = BookkeepingClient()
        self.__dataUsageClient = DataUsageClient()
        self.__fileCatalog = FileCatalog()
        self.cachedMetadata = {}
        # build a dictionary with Event Type descriptions (to be send to accounting, instead of number Event Type ID)
        self.eventTypeDescription = {
            "na": "na",
            "notInBkk": "notInBkk",
            "FailedBkkQuery": "FailedBkkQuery",
            "None": "None",
        }
        self.limitForCommit = self.am_getOption("LimitForCommit", 1000)
        self.callsToGetSummary = 0
        self.callsToDirectorySummary = 0
        self.callsToDudbForMetadata = 0
        # keep count of calls to Bkk
        self.callsToBkkgetDirectoryMetadata = 0
        self.callsToBkkGetEvtType = 0

        return S_OK()

    def userStorageAccounting(self):
        self.log.notice("-------------------------------------------------------------------------------------\n")
        self.log.notice("Generate accounting records for user directories ")
        self.log.notice("-------------------------------------------------------------------------------------\n")

        result = self.__stDB.getUserSummary()
        if not result["OK"]:
            return result
        userCatalogData = result["Value"]
        self.log.notice(f"Got summary for {len(userCatalogData)} users")
        result = self.__stDB.getUserSummaryPerSE()
        if not result["OK"]:
            return result
        userSEData = result["Value"]
        self.log.notice(f"Got SE summary for {len(userSEData)} users")

        now = datetime.datetime.utcnow()
        numRows = 0
        for user in sorted(userSEData):
            if user not in userCatalogData:
                self.log.error("User has SE data but not Catalog data!", user)
                continue
            for se in sorted(userSEData[user]):
                seData = userSEData[user][se]
                usRecord = UserStorage()
                usRecord.setStartTime(now)
                usRecord.setEndTime(now)
                usRecord.setValueByKey("User", user)
                usRecord.setValueByKey("StorageElement", se)
                usRecord.setValueByKey("LogicalSize", userCatalogData[user]["Size"])
                usRecord.setValueByKey("LogicalFiles", userCatalogData[user]["Files"])
                usRecord.setValueByKey("PhysicalSize", seData["Size"])
                usRecord.setValueByKey("PhysicalFiles", seData["Files"])
                usRecord.setValueByKey("StorageSize", 0)
                usRecord.setValueByKey("StorageFiles", 0)
                gDataStoreClient.addRegister(usRecord)
                numRows += 1

            self.log.notice(
                " User %s is using %.2f GiB (%s files)"
                % (user, userCatalogData[user]["Size"] / (1024.0**3), userCatalogData[user]["Files"])
            )
        self.log.notice(f"Sending {numRows} records to accounting for user storage")
        res = gDataStoreClient.commit()
        if not res["OK"]:
            self.log.notice(f"ERROR: committing UserStorage records: {res} ")
            return S_ERROR(res)
        else:
            self.log.notice(f"{numRows} records for UserStorage type successfully committed")

    def __getTopDirUsage(self, dirToScan, topDirLogicalUsage):
        """
        Get logical storage information from su_Directory table
        """
        res = self.__fileCatalog.listDirectory(dirToScan)
        if not res["OK"]:
            return res
        subDirs = res["Value"]["Successful"][dirToScan]["SubDirs"]
        doneDirs = set()
        # Get summary for each next level directory in order to reduce the size
        for directory in subDirs:
            # For  top directories, go one level below to reduce the size of information
            if len(directory.split(os.path.sep)) == 3:
                self.__getTopDirUsage(directory, topDirLogicalUsage)
                topDir = _standardDirectory(directory)
            elif directory.lower() not in doneDirs:
                # The StorageUsage DB is case insensitive, hence go only once over the same directory
                doneDirs.add(directory.lower())
                directory = _standardDirectory(directory)
                # get info from the DB about the LOGICAL STORAGE USAGE (from the su_Directory table):
                result = self.__stDB.getSummary(directory)
                if not result["OK"]:
                    return result
                logicalUsage = result["Value"]
                # Store logical usage for top level directory onl
                topDir = _standardDirectory(os.path.join(os.path.sep, *directory.split(os.path.sep)[:3]))
                for row in logicalUsage:
                    # d, size, files = row
                    topDirLogicalUsage[topDir]["Files"] += logicalUsage[row]["Files"]
                    topDirLogicalUsage[topDir]["Size"] += logicalUsage[row]["Size"]
            self.log.debug(
                f"After scan of {directory}, total of {topDir}: ",
                "size: %.4f TB  files: %d"
                % (
                    convertSizeUnits(topDirLogicalUsage[topDir]["Size"], "B", "TB"),
                    topDirLogicalUsage[topDir]["Files"],
                ),
            )
        return S_OK()

    def topDirectoryAccounting(self):
        """
        Get statistics for top level 2 directories, e.g. /lhcb/data/, /lhcb/user/ etc...
        """
        self.log.notice("-------------------------------------------------------------------------------------\n")
        self.log.notice("Generate accounting records for top directories ")
        self.log.notice("-------------------------------------------------------------------------------------\n")

        # Get second level top directories
        topDirLogicalUsage = defaultdict(lambda: {"Files": 0, "Size": 0})  # build the list of first level directories

        res = self.__getTopDirUsage("/lhcb", topDirLogicalUsage)
        if not res["OK"]:
            self.log.error("Error getting top directory logical usage", res["Message"])
            return res
        self.log.notice("Summary on logical usage of top directories: ")
        for topDir in topDirLogicalUsage:
            self.log.notice(
                "dir: %s size: %.4f TB  files: %d"
                % (
                    topDir,
                    convertSizeUnits(topDirLogicalUsage[topDir]["Size"], "B", "TB"),
                    topDirLogicalUsage[topDir]["Files"],
                )
            )

        # loop on top level directories (/lhcb/data/, /lhcb/user/, /lhcb/MC/, etc..)
        # to get the summary in terms of PHYSICAL usage grouped by SE:
        seData = {}
        for directory in topDirLogicalUsage:
            result = self.__stDB.getDirectorySummaryPerSE(directory)  # retrieve the PHYSICAL usage
            if not result["OK"]:
                return result
            seData[directory] = result["Value"]
            self.log.notice(f"Got SE summary for {len(seData)} directories ")
            self.log.debug(f"SEData: {seData}")
        # loop on top level directories to send the accounting records
        numRows = 0
        now = datetime.datetime.utcnow()
        for directory in seData:
            self.log.debug(f"dir: {directory} SEData: {seData[directory]} ")
            if directory not in topDirLogicalUsage:
                self.log.error(f"Dir {directory} is in the summary per SE, but it is not in the logical files summary!")
                continue
            for se in sorted(seData[directory]):
                storageRecord = Storage()
                storageRecord.setStartTime(now)
                storageRecord.setEndTime(now)
                storageRecord.setValueByKey("Directory", directory)
                storageRecord.setValueByKey("StorageElement", se)
                storageRecord.setValueByKey("LogicalFiles", topDirLogicalUsage[directory]["Files"])
                storageRecord.setValueByKey("LogicalSize", topDirLogicalUsage[directory]["Size"])
                try:
                    physicalFiles = seData[directory][se]["Files"]
                except Exception:
                    self.log.error(f"WARNING! no files replicas for directory {directory} on SE {se}")
                    physicalFiles = 0
                try:
                    physicalSize = seData[directory][se]["Size"]
                except Exception:
                    self.log.error(f"WARNING! no size for replicas for directory {directory} on SE {se}")
                    physicalSize = 0
                storageRecord.setValueByKey("PhysicalFiles", physicalFiles)
                storageRecord.setValueByKey("PhysicalSize", physicalSize)
                gDataStoreClient.addRegister(storageRecord)
                numRows += 1
                self.log.debug(
                    "Directory: %s SE: %s  physical size: %.4f TB (%d files)"
                    % (directory, se, convertSizeUnits(physicalSize, "B", "TB"), physicalFiles)
                )

        self.log.notice(f"Sending {numRows} records to accounting for top level directories storage")
        res = gDataStoreClient.commit()
        if not res["OK"]:
            self.log.notice(f"ERROR: committing Storage records: {res} ")
            return S_ERROR(res)
        else:
            self.log.notice(f"{numRows} records for Storage type successfully committed")

    def bkPathAccounting(self):
        """
        Generate accounting for all directories, grouped by BK path metadata
        """
        self.log.notice("-------------------------------------------------------------------------------------\n")
        self.log.notice("Generate accounting records for DataStorage type ")
        self.log.notice("-------------------------------------------------------------------------------------\n")

        # counter for DataStorage records, commit to the accounting in bunches of self.limitForCommit records
        self.totalRecords = 0
        self.recordsToCommit = 0
        self.log.notice(" Call the function to extract information from the StorageUsageDB..")
        res = self.generateStorageUsagePerDir()
        if not res["OK"]:
            self.log.error("ERROR querying the StorageUsageDB per directory")
            return S_ERROR()

        # Keep a list of all directories in FC that are not found in the Bkk
        self.directoriesNotInBkk = []
        # for debugging purposes build dictionaries with storage usage to compare with the accounting plots
        self.debug_seUsage = defaultdict(lambda: {"Files": 0, "Size": 0})
        self.debug_seUsage_acc = defaultdict(lambda: {"Files": 0, "Size": 0})

        # set the time for the accounting records (same time for all records)
        now = datetime.datetime.utcnow()
        # Get the directory metadata in a bulk query
        metaForList = self.__getMetadataForAcc(self.dirDict.values())

        # loop on all directories  to get the bkk metadata
        for dirLfn, fullDirectory in self.dirDict.items():
            if dirLfn not in fullDirectory:
                self.log.error(f"ERROR: fullDirectory should include the dirname: {fullDirectory} {dirLfn} ")
                continue
            self.log.debug(f"Processing directory {dirLfn} ")
            if dirLfn not in self.pfnUsage:
                self.log.error(f"ERROR: directory does not have PFN usage {dirLfn} ")
                continue
            self.log.debug(f"PFN usage: {self.pfnUsage[dirLfn]} ")
            if dirLfn not in self.lfnUsage:
                self.log.error(f"ERROR: directory does not have LFN usage {dirLfn} ")
                continue
            self.log.debug(f"LFN usage: {self.lfnUsage[dirLfn]} ")

            # for DEBUGGING:
            for se in self.pfnUsage[dirLfn]:
                self.debug_seUsage[se]["Files"] += self.pfnUsage[dirLfn][se]["Files"]
                self.debug_seUsage[se]["Size"] += self.pfnUsage[dirLfn][se]["Size"]
            # end of DEBUGGING

            # get metadata for this directory
            metaForDir = metaForList.get(fullDirectory, {})
            if not metaForDir:
                self.log.warn(f"Metadata not available for directory {fullDirectory}")
                continue

            # Fill in the accounting record
            self.log.info(f"Fill the record for {dirLfn} and metadata: {metaForDir} ")
            res = self.fillAndSendAccountingRecord(dirLfn, metaForDir, now)
            if not res["OK"]:
                return res
            for se in self.pfnUsage[dirLfn]:
                self.debug_seUsage_acc[se]["Files"] += self.pfnUsage[dirLfn][se]["Files"]
                self.debug_seUsage_acc[se]["Size"] += self.pfnUsage[dirLfn][se]["Size"]

        # Don't forget to commit the remaining records!
        self.__commitRecords()

    def execute(self):
        if self.am_getOption("CleanBefore", False):
            self.log.notice("Cleaning the DB")
            result = self.__stDB.purgeOutdatedEntries("/lhcb/user", self.am_getOption("OutdatedSeconds", 86400 * 10))
            if not result["OK"]:
                return result
            self.log.notice(f"Purged {result['Value']} outdated records")

        # User accounting (per user and SE)
        self.userStorageAccounting()
        # Accounting per top directory
        self.topDirectoryAccounting()
        # full production data accounting
        self.bkPathAccounting()

        self.log.notice("-------------------------------------------------------------------------------------\n")
        self.log.notice("------ End of cycle report for DataStorage accounting--------------------------------\n")
        self.log.notice(f"Total directories found in FC:  {len(self.dirDict)} ")
        totalCallsToStorageUsage = self.callsToGetSummary + self.callsToDirectorySummary
        self.log.notice("Total calls to StorageUsage: %d , took: %d s " % (totalCallsToStorageUsage, self.genTotalTime))
        totalCallsToBkk = self.callsToBkkgetDirectoryMetadata + self.callsToBkkGetEvtType
        self.log.notice("Total calls to DataUsage for cache: %d" % self.callsToDudbForMetadata)
        self.log.notice(
            "Total calls to Bookkeeping: %d (getDirectoryMetadata: %d, getEventType: %d)"
            % (totalCallsToBkk, self.callsToBkkgetDirectoryMetadata, self.callsToBkkGetEvtType)
        )
        self.log.notice("Total records sent to accounting for DataStorage:  %d " % self.totalRecords)
        self.log.notice(f"Directories not found in Bookkeeping: {len(self.directoriesNotInBkk)} ")
        fileName = os.path.join(self.__workDirectory, "directoriesNotInBkk.txt")
        self.log.notice(f"written to file: {fileName} ")
        fd = open(fileName, "w")
        for dd in self.directoriesNotInBkk:
            fd.write(f"{dd}\n")
        fd.close()
        # for DEBUG only
        self.log.info("Summary of StorageUsage: files size ")
        for se in sorted(self.debug_seUsage):
            self.log.info(
                "all: %s  %d %d Bytes ( %.2f TB ) "
                % (
                    se,
                    self.debug_seUsage[se]["Files"],
                    self.debug_seUsage[se]["Size"],
                    self.debug_seUsage[se]["Size"] / 1.0e12,
                )
            )
            if se in self.debug_seUsage_acc:
                self.log.info(
                    "acc: %s  %d %d Bytes ( %.2f TB ) "
                    % (
                        se,
                        self.debug_seUsage_acc[se]["Files"],
                        self.debug_seUsage_acc[se]["Size"],
                        self.debug_seUsage_acc[se]["Size"] / 1.0e12,
                    )
                )
            else:
                self.log.info("SE not in self.debug_seUsage_acc keys")
        return S_OK()

    def __getMetadataForAcc(self, dirList):
        """Get metadata for a directory either from memory, from the storageDB or
        from BK."""
        # Try and get the metadata from memory cache
        notFound = []
        metaForList = {}
        for dirName in dirList:
            metaForList[dirName] = self.cachedMetadata.get(dirName, {})
            if not metaForList[dirName]:
                notFound.append(dirName)
        notInCache = []
        if notFound:
            self.log.info(f"Memory metadata cache missed for {len(notFound)} directories")
            self.log.debug(f"call getDirMetadata for (first 10): {str(notFound[0:10])} ")
            for dirChunk in breakListIntoChunks(notFound, 10000):
                self.callsToDudbForMetadata += 1
                res = self.__dataUsageClient.getDirMetadata(
                    dirChunk
                )  # this could be a bulk query for a list of directories
                if not res["OK"]:
                    self.log.error(f"Error retrieving {len(dirChunk)} directories meta-data {res['Message']} ")
                    # this usually happens when directories are removed from FC between the StorageUsageDB dump and this call,
                    # if the Db is refreshed exactly in this time interval. Not really a problem.
                    # 3 just a try ##############################################3
                    notInCache += dirChunk
                    continue
                self.log.debug(f"getDirMetadata returned: {str(res['Value'])} ")
                for dirName in dirChunk:
                    # Compatibility with old (list for single file) and new (dictionary) service
                    if isinstance(res["Value"], type({})):
                        metaTuple = res["Value"].get(dirName, ())
                    elif len(dirList) == 1 and res["Value"]:
                        metaTuple = res["Value"][0]
                    else:
                        metaTuple = ()
                    if metaTuple and metaTuple[3] is not None:
                        metaForDir = metaForList[dirName]
                        (
                            _dirID,
                            metaForDir["DataType"],
                            metaForDir["Activity"],
                            metaForDir["Conditions"],
                            metaForDir["ProcessingPass"],
                            metaForDir["EventType"],
                            metaForDir["FileType"],
                            metaForDir["Production"],
                            metaForDir["Visibility"],
                        ) = metaTuple
                    else:
                        notInCache.append(dirName)

            failedBK = []
            if notInCache:
                cachedFromBK = []
                self.log.info(
                    f"Directory metadata cache missed for {len(notInCache)} directories => query BK and cache"
                )
                for dirChunk in breakListIntoChunks(notInCache, 200):
                    self.callsToBkkgetDirectoryMetadata += 1
                    res = self.__bkClient.getDirectoryMetadata(dirChunk)
                    if not res["OK"]:
                        self.log.error("Totally failed to query Bookkeeping", res["Message"])
                        failedBK += dirChunk
                        for dirName in dirChunk:
                            metaForDir = metaForList[dirName]
                            _fillMetadata(metaForDir, "FailedBkkQuery")
                    else:
                        bkMetadata = res["Value"]
                        self.log.debug(f"Successfully queried Bookkeeping, result: {bkMetadata} ")
                        for dirName in dirChunk:
                            metaForDir = metaForList[dirName]
                            # BK returns a list of metadata, chose the first one...
                            metadata = bkMetadata["Successful"].get(dirName, [{}])[0]
                            if metadata and metadata.get("ConditionDescription") is not None:
                                metadata["Visibility"] = metadata.pop(
                                    "VisibilityFlag", metadata.get("Visibility", "na")
                                )
                                # All is OK, directory found
                                _fillMetadata(metaForDir, metadata)
                                self.log.debug(f"Cache entry {dirName} in DirMetadata table..")
                                resInsert = self.__dataUsageClient.insertToDirMetadata({dirName: metadata})
                                if not resInsert["OK"]:
                                    self.log.error(
                                        "Failed to cache metadata:", f"{resInsert['Message']} for dir {dirName}"
                                    )
                                else:
                                    cachedFromBK.append(dirName)
                                    self.log.debug(f"Successfully cached metadata for {dirName} : {str(metadata)}")
                                    self.log.debug(f"result: {str(resInsert)} ")
                            else:
                                # Directory not found
                                self.log.debug(f"Directory {dirName} not registered in Bookkeeping!")
                                _fillMetadata(metaForDir, "notInBkk")
                                failedBK.append(dirName)
                                self.directoriesNotInBkk.append(dirName)
                            # Translate a few keys for accounting
                            for bkName, accName in (
                                ("ConfigName", "DataType"),
                                ("ConfigVersion", "Activity"),
                                ("ConditionDescription", "Conditions"),
                            ):
                                metaForDir[accName] = metaForDir.pop(bkName, "na")
                self.log.info(f"Successfully cached {len(cachedFromBK)} directories from BK")
                if self.directoriesNotInBkk:
                    self.log.warn(f"{len(self.directoriesNotInBkk)} directories not found in BK")

            # cache locally the metadata
            for dirName in [dn for dn in notFound if dn not in failedBK]:
                metaForDir = metaForList[dirName]
                # Translate the numerical event type to a description string
                metaForDir["EventType"] = self.__getEventTypeDescription(metaForDir.pop("EventType", "na"))
                self.cachedMetadata[dirName] = metaForDir.copy()
        else:
            self.log.info(f"Memory metadata cache hit for {len(dirList)} directories")
        return metaForList

    def __commitRecords(self):
        if self.recordsToCommit:
            res = gDataStoreClient.commit()
            if not res["OK"]:
                self.log.error(f"Accounting ERROR: commit returned {res}")
            else:
                self.log.notice("%d records committed " % self.recordsToCommit)
                self.recordsToCommit = 0
                self.log.notice(f"commit for DataStorage returned: {res}")

    def generateStorageUsagePerDir(self):
        """
        Extract storage info from the StorageUsageDB and keep it in memory in
        dictionaries self.lfnUsage and self.pfnUsage
        """

        start = time.time()
        self.log.notice("Starting from path: /lhcb/")
        res = self.__stDB.getStorageDirectories("/lhcb/")
        if not res["OK"]:
            return S_ERROR()
        totalDirList = res["Value"]
        self.log.info(f"Total directories retrieved from StorageUsageDB: {len(totalDirList)} ")
        # select only terminal directories (directories without sub-directories)
        # mc directory structure: /lhcb/MC/[year]/[file type]/[prod]/0000/ => len = 7
        # raw data:               /lhcb/data/2011/RAW/FULL/LHCb/COLLISION11/99983
        # => len 9 (under /lhcb/data/ from 2011 only RAW, before 2011 also other file types)
        # processed data: under both /lhcb/data and /lhcb/LHCb/
        #                         /lhcb/data/2010/DST/00009300/0000
        # data:                   /lhcb/LHCb/Collision12/ICHEP.DST/00017399/0000/
        self.dirDict = {}
        ignoredDirectories = dict.fromkeys(self.__ignoreDirsList, 0)
        self.log.info(f"Directories to be ignored: {str(sorted(ignoredDirectories))} ")
        for dirItem in totalDirList:
            # make sure that last character is a '/'
            dirItem = _standardDirectory(dirItem)
            splitDir = dirItem.split(os.path.sep)
            if len(splitDir) < 4:  # avoid picking up intermediate directories which don't contain files, like /lhcb/
                self.log.warn(f"Directory {dirItem} skipped, as top directory")
                continue
            secDir = splitDir[2]
            if secDir in ignoredDirectories:
                self.log.debug(f"Directory to be ignored, skipped: {dirItem} ")
                ignoredDirectories[secDir] += 1
                continue
            # for each type of directory (MC, reconstructed data and raw data) check the format,
            # in order not to count more than once the productions with more than one sub-directory
            # for MC directories:
            # example: '/lhcb/MC/MC10/ALLSTREAMS.DST/00010908/0000/',
            # or        /lhcb/MC/2011/DST/00010870/0000
            # one directory for each file type
            # for histograms, there is no numeric "suffix" like /0000/
            # example: /lhcb/LHCb/Ionproton13/HIST/136973/
            # for data
            # production: /lhcb/LHCb/Collision11/SWIMSTRIPPINGD02KSPIPI.MDST/00019088/0000/
            # for raw data: /lhcb/data/2012/RAW/FULL/LHCb/COLLISION12/133784/
            try:
                # RAW data directories have a special format, see above
                if splitDir[-6] == "RAW":
                    self.log.debug(f"RAW DATA directory: {splitDir}")
                    directory = dirItem
                    fullDirectory = directory
                else:
                    # These are production directories, see above for the format
                    self.log.debug(f"MC or reconstructed data directory: {splitDir}")
                    # HIST directories do not have a "suffix" (i.e. all in the same directory)!
                    if splitDir[-3] == "HIST":
                        directory = dirItem
                        fullDirectory = directory
                    else:
                        # Ignore the suffix for the accounting as this is only an artifact
                        directory = _standardDirectory(os.path.dirname(os.path.dirname(dirItem)))
                        fullDirectory = dirItem
                # Keep the link between the accounting directory and the full directory name
                self.dirDict.setdefault(directory, fullDirectory)
                self.log.debug(f"Directory contains production files: {directory} ")
            except Exception:
                self.log.warn(f"The directory has unexpected format: {dirItem} ")

        self.lfnUsage = defaultdict(lambda: {"LfnSize": 0, "LfnFiles": 0})
        self.pfnUsage = {}

        totalDiscardedDirs = 0
        self.log.info("Directories that have been discarded:")
        for dd in ignoredDirectories:
            self.log.info("/lhcb/%s - %d " % (dd, ignoredDirectories[dd]))
            totalDiscardedDirs += ignoredDirectories[dd]
        self.log.info("Total discarded directories: %d " % totalDiscardedDirs)

        self.log.info(f"Retrieved {len(self.dirDict)} dirs from StorageUsageDB containing prod files")
        self.log.info("Getting the number of files and size from StorageUsage service")

        for directory in self.dirDict:
            self.log.debug(f"Get storage usage for directory {directory} ")
            res = self.__stDB.getDirectorySummaryPerSE(directory)
            self.callsToDirectorySummary += 1
            if not res["OK"]:
                self.log.error("Cannot retrieve PFN usage", res["Message"])
                continue
            # save the PFN usage per SE in a dictionary
            self.pfnUsage.setdefault(directory, res["Value"])

            self.log.debug(f"Get logical usage for directory {directory} ")
            # This returns the number of files and size for the directory and all its subdirectories
            res = self.__stDB.getSummary(directory)
            self.callsToGetSummary += 1
            if not res["OK"]:
                self.log.error("Cannot retrieve LFN usage", res["Message"])
                continue
            if not res["Value"]:
                self.log.error("getSummary returned empty value", f"for {directory}: {str(res)}")
                continue
            # Sum up all subdirectories
            for dirInfo in res["Value"].values():  # can be an iterator
                self.lfnUsage[directory]["LfnSize"] += dirInfo["Size"]
                self.lfnUsage[directory]["LfnFiles"] += dirInfo["Files"]
            self.log.debug(f"PFN usage: {self.pfnUsage[directory]}")
            self.log.debug(f"LFN usage: {self.lfnUsage[directory]}")

        end = time.time()
        self.genTotalTime = end - start
        self.log.info("StorageUsageDB extraction completed in %d s" % self.genTotalTime)

        return S_OK()

    def __getEventTypeDescription(self, eventType):
        # convert eventType to string:
        try:
            eventType = int(eventType)
        except Exception:
            pass
        # check that the event type description is in the cached dictionary, and otherwise query the Bkk
        if eventType not in self.eventTypeDescription:
            self.log.notice(f"Event type description not available for eventTypeID {eventType}, getting from Bkk")
            res = self.__bkClient.getAvailableEventTypes()
            self.callsToBkkGetEvtType += 1
            if not res["OK"]:
                self.log.error("Error querying the Bkk:", res["Message"])
            else:
                self.eventTypeDescription.update(dict(res["Value"]))
            self.log.debug(f"Updated  self.eventTypeDescription dict: {str(self.eventTypeDescription)} ")
            # If still not found, log it!
            if eventType not in self.eventTypeDescription:
                self.log.error(f"EventType {str(eventType)} is not in cached dictionary")

        return self.eventTypeDescription.get(eventType, "na")

    def fillAndSendAccountingRecord(self, lfnDir, metadataDict, now):
        """Create, fill and send to accounting a record for the DataStorage
        type."""
        dataRecord = DataStorage()
        dataRecord.setStartTime(now)
        dataRecord.setEndTime(now)
        logicalSize = self.lfnUsage[lfnDir]["LfnSize"]
        logicalFiles = self.lfnUsage[lfnDir]["LfnFiles"]
        dataRecord.setValueByKey("LogicalSize", logicalSize)
        dataRecord.setValueByKey("LogicalFiles", logicalFiles)
        for key in ("DataType", "Activity", "FileType", "Production", "ProcessingPass", "Conditions", "EventType"):
            dataRecord.setValueByKey(key, metadataDict.get(key, "na"))
        self.log.debug(">>> Send DataStorage record to accounting:")
        self.log.debug("\tlfnFiles: %d lfnSize: %d " % (logicalFiles, logicalSize))

        for se in self.pfnUsage[lfnDir]:
            self.log.debug(f"Filling accounting record for se {se}")
            physicalSize = self.pfnUsage[lfnDir][se]["Size"]
            physicalFiles = self.pfnUsage[lfnDir][se]["Files"]

            dataRecord.setValueByKey("StorageElement", se)
            dataRecord.setValueByKey("PhysicalSize", physicalSize)
            dataRecord.setValueByKey("PhysicalFiles", physicalFiles)
            self.log.debug(
                "\t\tStorageElement: %s --> physFiles: %d  physSize: %d " % (se, physicalFiles, physicalSize)
            )

            # addRegister is NOT making a copy, therefore all records are otherwise overwritten
            res = gDataStoreClient.addRegister(copy.deepcopy(dataRecord))
            if not res["OK"]:
                self.log.error(f"addRegister returned: {res}")
                return S_ERROR(f"addRegister returned: {res}")
            # Reset logical information to zero in order to send it only once!
            dataRecord.setValueByKey("LogicalSize", 0)
            dataRecord.setValueByKey("LogicalFiles", 0)
            self.totalRecords += 1
            self.recordsToCommit += 1

        # Commit if necessary
        if self.recordsToCommit > self.limitForCommit:
            self.__commitRecords()

        return S_OK()



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import json
import os
import pytest

from LHCbDIRAC.BookkeepingSystem.DB.DataTakingConditionInterpreter import generateConditionDescription


with open(os.path.join(os.path.dirname(__file__), "dtc.json")) as f:
    allExistingConditions = json.load(f)


KNOWN_TO_DIFFER = {
    "1019-LHCb": "VeloOpen-MagOff-Excl-EC-HC-IT-MU-OT-R1-SP-TT-VE",
    "1019-RICH2": "MagOff",
    "1019-RICH": "MagOff",
    "1024-RICH1": "MagOff",
    "1024-RICH2": "MagOff",
    "1024-RICH": "MagOff",
    "1378-LHCb": "VeloOpen-MagOff-Excl-IT-MU-OT-R1-R2-TT-VE",
    "1516-LHCb": "VeloOpen-MagOff-Excl-EC-HC-IT-MU-OT-R1-R2-SP-TT-VE",
    "1516-MUON": "MagOff",
    "1516-RICH2": "MagOff",
    "1516-TDET": "MagOff",
    "1516-VELOA": "MagOff",
    "1516-VELOC": "MagOff",
    "1516-VELO": "VeloOpen-MagOff",
    "1583-LHCb": "VeloOpen-MagOff-Excl-IT-OT-R1-R2-TT-VE",
    "182-CALO": "MagOff",
    "182-ECAL": "MagOff",
    "182-HCAL": "MagOff",
    "182-HRC": "MagOff",
    "182-IT": "MagOff",
    "182-LHCb": "VeloOpen-MagOff-Excl-EC-HC-MU-OT-R1-R2-SP-TT-VE",
    "182-MUON": "MagOff",
    "182-OT": "MagOff",
    "182-PRS": "MagOff",
    "182-RICH1": "MagOff",
    "182-RICH": "MagOff",
    "182-TDET": "MagOff",
    "182-TT": "MagOff",
    "182-VELOC": "MagOff",
    "200-LHCb": "BeamOff-VeloOpen-MagOff-Excl-EC-HC-IT-MU-OT-R1-R2-SP-TT-VE",
    "200-VELO": "BeamOff-VeloOpen-MagOff",
    "267-LHCb": "VeloOpen-MagOff-Excl-OT-R1-R2-VE",
    "366-RICH1": "BeamOff-MagOff",
    "366-RICH": "BeamOff-MagOff",
    "428095-OT": "BeamOff-MagOff",
    "429155-certification": "Beam3500GeV-MagDown",
    "429155-MC": "Beam3500GeV-MagDown",
    "429155-test": "Beam3500GeV-MagDown",
    "429155-validation": "Beam3500GeV-MagDown",
    "429215-test": "Beam3500GeV-MagUp",
    "429215-validation": "Beam3500GeV-MagUp",
    "429695-FEST": "Beam4000GeV-MagDown",
    "429695-validation": "Beam4000GeV-MagDown",
    "430455-validation": "Beam1380GeV-MagDown",
    "431576-validation": "Beam4000GeV-MagUp",
    "432178-validation": "Beam1380GeV-MagDown",
    "432900-validation": "Beam6500GeV-MagUp",
    "432943-validation": "Beam6500GeV-MagDown",
    "433198-validation": "Beam2510GeV-MagDown",
    "433204-validation": "Beam6370GeV-MagDown",
    "4937-IT": "BeamOff-MagOff",
    "5432-RICH2": "BeamOff-MagOff",
    "5432-RICH": "BeamOff-MagOff",
    "5476-LHCb": "VeloOpen-MagOff-Excl-HC-IT-MU-OT-R1-R2-TT-VE",
    "5687-RICH2": "MagOff",
    "5811-LHCb": "VeloOpen-MagOff-Excl-EC-HC-IT-OT-R1-R2-SP-TT-VE",
    "6136-Fest": "BeamOff-MagOff",
    "6136-FEST": "BeamOff-MagOff",
    "6138-LHCb": "BeamOff-VeloOpen-MagOff-Excl-EC-HC-IT-MU-OT-R1-R2-SP-TT-VE",
    "849-LHCb": "VeloOpen-MagOff-Excl-EC-HC-IT-MU-OT-R1-R2-SP-VE",
    "849-TT": "MagOff",
    "880-RICH2": "MagOff",
    "880-RICH": "MagOff",
    # There was a very special case for when the config name was VELO
    # in which case the Velo position was in the description, as opposed
    # to any other string. Forget about that use case.
    "428095-VELO": "BeamOff-VeloOpen-MagOff",
    "428116-VELO": "BeamOff-VeloOpen-MagOff",
    "6136-VELO": "BeamOff-VeloOpen-MagOff",
}


@pytest.mark.parametrize("dataTakingDict", allExistingConditions)
def test_existingConditions(dataTakingDict):
    """This checks whether the condition descriptions stored
    in the DB can still be generated from the list of sub-detector.
    We know that it is not the case for some of them, even before changing code
    """

    # The json file was obtained by dumping the production DB:
    #
    # select distinct ConfigName, dc.* \
    # from data_taking_conditions dc\
    # JOIN PRODUCTIONSCONTAINER pc on dc.DAQPERIODID = pc.DAQPERIODID \
    # JOIN CONFIGURATIONS c on c.CONFIGURATIONID=pc.CONFIGURATIONID;
    #
    # and then edit it
    #
    # sed -e 's/"it"/"IT"/g' -e 's/"tt"/"TT"/g' -e 's/"ot"/"OT"/g' \
    #     -e 's/"rich1"/"RICH1"/g' -e 's/"rich2"/"RICH2"/g' \
    #     -e 's/spd_prs/SPD_PRS/g' -e 's/"ecal"/"ECAL"/g' \
    #     -e 's/"hcal"/"HCAL"/g' -e 's/"muon"/"MUON"/g' -e 's/"l0"/"L0"/g' \
    #     -e 's/"hlt"/"HLT"/g' -e 's/"veloposition"/"VeloPosition"/g' \
    #     -e 's/configname/ConfigName/g' -e 's/configversion/ConfigVersion/g' \
    #     -e 's/description/Description/g' -e 's/beamcond/BeamCond/g' \
    #     -e 's/beamenergy/BeamEnergy/g' -e 's/magneticfield/MagneticField/g' \
    #     -e 's/"velo"/"VELO"/g' -e 's/configname/ConfigName/g' \
    #     -e 's/configversion/ConfigVersion/g' -e 's/description/Description/g' \
    #     -e 's/beamcond/BeamCond/g' -e 's/beamenergy/BeamEnergy/g' \
    #     -e 's/magneticfield/MagneticField/g' -e 's/"velo"/"VELO"/g'

    configName = dataTakingDict.pop("ConfigName")

    descriptionHash = f"{dataTakingDict['daqperiodid']}-{configName}"

    if descriptionHash in KNOWN_TO_DIFFER:
        pytest.xfail("Known to differ, even in the old system")
    else:
        description = generateConditionDescription(dataTakingDict, configName)
        assert (
            description == dataTakingDict["Description"]
        ), f"{description}, {dataTakingDict['Description']}, {descriptionHash}"
